# Encoding: utf-8
"""Dispersion specified by a table of wavelengths (nm) and dielectric function values."""

from typing import Any, Dict, Union
import numpy.typing as npt
import scipy.interpolate

from .epsilon_inf import EpsilonInf
from .base_dispersion import Dispersion, InvalidParameters, DispersionSum


class TableEpsilon(Dispersion):
    r"""Dispersion specified by a table of wavelengths (nm) and dielectric function values.
    Please not that this model will produce errors for wavelengths outside the provided
    wavelength range.

    Single parameters:
        :lbda (list): Wavelengths in nm. This value must be provided.
        :epsilon: Complex dielectric function values in the convention ε1 + iε2.
            This value must be provided.
        :kind: Type of interpolation
            (see scipy.interpolate.interp1d for more information). Defaults to 'linear'.

    Repeated parameters:
        --

    Output:
        The interpolation in the given wavelength range.
    """

    single_params_template = {"lbda": None, "epsilon": None}
    rep_params_template: Dict[str, Any] = {}

    def __init__(self, *args, **kwargs) -> None:
        self.kind = kwargs.pop("kind", "linear")

        super().__init__(*args, **kwargs)

        if len(self.single_params.get("lbda")) == 0:
            raise InvalidParameters("Wavelength array cannot be of length zero.")

        if len(self.single_params.get("epsilon")) != len(
            self.single_params.get("lbda")
        ):
            raise InvalidParameters(
                "Wavelength and epsilon arrays must have the same length."
            )

        self.interpolation = scipy.interpolate.interp1d(
            self.single_params.get("lbda"),
            self.single_params.get("epsilon"),
            kind=self.kind,
        )

        self.default_lbda_range = self.single_params.get("lbda")

    def __add__(self, other: Union[int, float, "Dispersion"]) -> "DispersionSum":
        if isinstance(other, (int, float)):
            return DispersionSum(self, EpsilonInf(eps=other))

        if isinstance(other, TableEpsilon):
            raise NotImplementedError(
                "Adding of tabular dispersions is not yet supported"
            )

        if isinstance(other, Dispersion):
            return DispersionSum(self, other)

        if isinstance(other, DispersionSum):
            other.dispersions.append(self)
            return other

        raise TypeError(
            f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'"
        )

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return self.interpolation(lbda)



"""Benchmark for using the formula dispersion"""

from pytest import fixture
import numpy as np

import elli
from elli.fitting import ParamsHist


wavelength = np.linspace(400, 800, 500)
PHI = 70


@fixture
def structure():
    """Build a structure with a formula dispersion"""
    params = ParamsHist()
    params.add("SiO2_n0", value=1.452, min=-100, max=100, vary=False)
    params.add("SiO2_n1", value=36.0, min=-40000, max=40000, vary=False)
    params.add("SiO2_n2", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_k0", value=0, min=-100, max=100, vary=False)
    params.add("SiO2_k1", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_k2", value=0, min=-40000, max=40000, vary=False)
    params.add("SiO2_d", value=276.36, min=0, max=40000, vary=False)

    params.add("TiO2_n0", value=2.236, min=-100, max=100, vary=True)
    params.add("TiO2_n1", value=451, min=-40000, max=40000, vary=True)
    params.add("TiO2_n2", value=251, min=-40000, max=40000, vary=True)
    params.add("TiO2_k0", value=0, min=-100, max=100, vary=False)
    params.add("TiO2_k1", value=0, min=-40000, max=40000, vary=False)
    params.add("TiO2_k2", value=0, min=-40000, max=40000, vary=False)

    params.add("TiO2_d", value=20, min=0, max=40000, vary=True)

    SiO2 = elli.FormulaIndex(
        "n = n0 + 1e2 * n1 / lbda ** 2 + 1e7 * n2 / lbda ** 4 + "
        "1j * (k0 + 1e2 * k1 / lbda ** 2 + 1e7 * k2 / lbda ** 4)",
        "lbda",
        {
            "n0": params["SiO2_n0"].value,
            "n1": params["SiO2_n1"].value,
            "n2": params["SiO2_n2"].value,
            "k0": params["SiO2_k0"].value,
            "k1": params["SiO2_k1"].value,
            "k2": params["SiO2_k2"].value,
        },
        {},
        "nm",
    ).get_mat()
    TiO2 = elli.FormulaIndex(
        "n = n0 + 1e2 * n1 / lbda ** 2 + 1e7 * n2 / lbda ** 4 + "
        "1j * (k0 + 1e2 * k1 / lbda ** 2 + 1e7 * k2 / lbda ** 4)",
        "lbda",
        {
            "n0": params["TiO2_n0"].value,
            "n1": params["TiO2_n1"].value,
            "n2": params["TiO2_n2"].value,
            "k0": params["TiO2_k0"].value,
            "k1": params["TiO2_k1"].value,
            "k2": params["TiO2_k2"].value,
        },
        {},
        "nm",
    ).get_mat()

    Layer = [
        elli.Layer(TiO2, params["TiO2_d"]),
        elli.Layer(SiO2, params["SiO2_d"]),
        elli.Layer(TiO2, params["TiO2_d"]),
        elli.Layer(SiO2, params["SiO2_d"]),
        elli.Layer(TiO2, params["TiO2_d"]),
        elli.Layer(SiO2, params["SiO2_d"]),
        elli.Layer(TiO2, params["TiO2_d"]),
        elli.Layer(SiO2, params["SiO2_d"]),
    ]

    return elli.Structure(elli.AIR, Layer, elli.AIR)


def test_formula_solver2x2(benchmark, structure):
    """Benchmarks solver2x2"""
    benchmark.pedantic(
        structure.evaluate,
        args=(wavelength, PHI),
        kwargs={"solver": elli.Solver2x2},
        iterations=1,
        rounds=10,
    )


def test_formula_solver4x4_expm(benchmark, structure):
    """Benchmarks solver2x2"""
    benchmark.pedantic(
        structure.evaluate,
        args=(wavelength, PHI),
        kwargs={"solver": elli.Solver4x4, "propagator": elli.PropagatorExpm()},
        iterations=1,
        rounds=10,
    )



# Encoding: utf-8
r"""The experiment class defines the external conditions of the sample.

It requires a Structure to evaluate and includes the information about
the incident light beam:

* the wavelengths :math:`\lambda`
* the incidence angle :math:`\theta_\text{i}`
* and the polarization, which can be given by a Jones or Stokes vector

The evaluate method can be called, to start the calculation of the optical properties.
To choose a Solver to be used in the calculation, the solver class is provided
as an argument and an object will be created automatically.

The experiment class is only needed in special cases an can be skipped by calling
:meth:`elli.structure.Structure.evaluate`.
"""

import numpy as np
import numpy.typing as npt

from .result import Result
from .solver import Solver
from .solver4x4 import Solver4x4


class Experiment:
    """Description of a virtual experiment to simulate the behavior of a structure."""

    structure = None
    jones_vector = None
    stokes_vector = None
    theta_i = None
    lbda = None

    def __init__(
        self,
        structure: "Structure",
        lbda: npt.ArrayLike,
        theta_i: float,
        vector: npt.ArrayLike = None,
    ) -> None:
        """Creates a virtual experiment to simulate the behavior of a structure.

        Args:
            structure (Structure): Structure object to evaluate.
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).
            theta_i (float): Incident angle (in degrees).
            vector (npt.ArrayLike, optional):
                Jones or Stokes vector of incident light. Defaults to diagonal polarization ([1, 0, 1, 0]).
        """
        self.set_structure(structure)
        self.set_theta(theta_i)
        self.set_lbda(lbda)
        self.set_vector(vector)

    def set_structure(self, structure: "Structure") -> None:
        """Defines the Structure to evaluate.

        Args:
            structure (Structure): Structure object to evaluate.
        """
        self.structure = structure

    def set_vector(self, vector: npt.ArrayLike) -> None:
        """Defines the Jones or Stokes vector of the incident Light.

        Jones:
        [1, 0]: horizontal polarized
        [0, 1]: vertical polarized
        [1/sqrt(2), 1/sqrt(2)]: diagonal polarized
        [1/sqrt(2),-1/sqrt(2)]: anti-diagonal polarized

        Stokes:
        [1,0,0,0]: unpolarized light
        [1,1,0,0]: horizontal polarized
        [1,-1,0,0]: vertical polarized
        [1,0,1,0]: diagonal polarized
        [1,0,-1,0]: anti-diagonal polarized

        Args:
            vector (npt.ArrayLike): Jones or Stokes vector of incident light.
        """
        if vector is None:
            vector = [1, 0, 1, 0]

        vector = np.asarray(vector)

        if vector.shape == (2,):
            self.jones_vector = vector

            self.stokes_vector = np.array(
                [
                    np.abs(self.jones_vector[0]) ** 2
                    + np.abs(self.jones_vector[1]) ** 2,
                    np.abs(self.jones_vector[0]) ** 2
                    - np.abs(self.jones_vector[1]) ** 2,
                    2
                    * np.real(
                        self.jones_vector[0] * np.conjugate(self.jones_vector[1])
                    ),
                    -2
                    * np.imag(
                        self.jones_vector[0] * np.conjugate(self.jones_vector[1])
                    ),
                ]
            )

        elif vector.shape == (4,):
            self.stokes_vector = vector

            p = (
                np.sqrt(
                    self.stokes_vector[1] ** 2
                    + self.stokes_vector[2] ** 2
                    + self.stokes_vector[3] ** 2
                )
                / self.stokes_vector[0]
            )
            Q = self.stokes_vector[1] / (self.stokes_vector[0] * p)
            U = self.stokes_vector[2] / (self.stokes_vector[0] * p)
            V = self.stokes_vector[3] / (self.stokes_vector[0] * p)

            if Q == -1:
                a = 0
                b = 1
            else:
                a = np.sqrt((1 + Q) / 2)
                b = U / (2 * a) - 1j * V / (2 * a)

            self.jones_vector = np.array([a, b])

    def set_theta(self, theta_i: float) -> None:
        """Set incident angle to evaluate.

        Args:
            theta_i (float): Incident angle (in degrees).
        """
        self.theta_i = theta_i

    def set_lbda(self, lbda: npt.ArrayLike) -> None:
        """Set experiment wavelengths.

        Args:
            lbda (npt.ArrayLike): single value or array of wavelengths (in nm).
        """
        lbda_array = np.asarray(lbda)
        if np.shape(lbda_array) == ():
            lbda_array = np.asarray([lbda])
        self.lbda = lbda_array

    def evaluate(self, solver: Solver = Solver4x4, **solver_kwargs) -> Result:
        """Evaluates the experiment with the given solver.

        Args:
            solver (Solver, optional): Choose which solver class is used. Defaults to Solver4x4.
            solver_kwargs (optional): Keyword arguments for the Solver can be appended as arguments.

        Returns:
            Result: Result of the experiment.
        """
        solv = solver(self, **solver_kwargs)
        return solv.calculate()



import os
from shutil import copytree, rmtree

from pytest import fixture


@fixture
def datadir(tmp_path, request):
    """
    Fixture responsible for searching a folder with the same name of test
    module and, if available, moving all contents to a temporary directory so
    tests can use them freely.
    """
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)

    if os.path.isdir(test_dir):
        rmtree(tmp_path)
        copytree(test_dir, str(tmp_path))

    return tmp_path



# Encoding: utf-8
from .cauchy import Cauchy
from .cauchy_custom import CauchyCustomExponent
from .cauchy_urbach import CauchyUrbach
from .constant_refractive_index import ConstantRefractiveIndex
from .drude_energy import DrudeEnergy
from .drude_resistivity import DrudeResistivity
from .epsilon_inf import EpsilonInf
from .gaussian import Gaussian
from .lorentz_energy import LorentzEnergy
from .lorentz_lambda import LorentzLambda
from .poles import Poles
from .polynomial import Polynomial
from .sellmeier import Sellmeier
from .sellmeier_custom import SellmeierCustomExponent
from .table_epsilon import TableEpsilon
from .table_index import Table
from .table_spectraray import TableSpectraRay
from .tanguy import Tanguy
from .tauc_lorentz import TaucLorentz
from .cody_lorentz import CodyLorentz
from .pseudo_dielectric import PseudoDielectricFunction
from .formula import Formula, FormulaIndex



"""
A helper class to load data from Woollam ASCII Files.
It supports loading of standard psi/delta values.
"""

import logging
import re
from typing import TextIO

import pandas as pd
from pint import DimensionalityError, UndefinedUnitError

from ..units import ureg
from ..utils import calc_rho
from . import detect_encoding

logger = logging.getLogger(__name__)


is_float_regex = re.compile(r"[+-]?(\d+([.]\d*)?([eE][+-]?\d+)?|[.]\d+([eE][+-]?\d+)?)")


def is_float(line: str) -> bool:
    """Checks whether the given line is a float

    Args:
        line (str): The line presumably containing a float

    Returns:
        bool: True if it is a float, False otherwise
    """
    return bool(is_float_regex.search(line))


def _is_wvase_tabular(line: str) -> bool:
    """
    Checks whether the provided line is in wvase tabular layout.

    Args:
        line (str): The line to check

    Returns:
        bool: True if the line is in wvase tabular layout.
    """
    return bool(re.search(r"^(\d*\.\d*\s+){5}\d*\.\d*$", line))


def _is_complete_ease_tabular(line: str) -> bool:
    """
    Checks whether the provided line is in complete ease tabular layout.

    Args:
        line (str): The line to check

    Returns:
        bool: True if the line is in complete ease tabular layout.
    """
    return bool(re.search(r"^[a-zA-Z]+\s+(\d+\.\d+\s+){5}\d+\.\d+$", line))


def _is_tan_cos_format(line: str) -> bool:
    """
    Checks whether the line denotes a wvase tan/cos format.

    Args:
        line (str): The line to check

    Returns:
        bool: True if the line is of the format {unit} TRIG.
    """
    return bool(re.search(r"^[a-zA-Z]+\s+TRIG", line))


def scale_to_nm(unit: str, dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Scales the wavelength axis of the given dataframe from the
    provided unit to nm.

    Args:
        unit (str): The unit string. Should be pint compatible.
        dataframe (pd.DataFrame): The dataframe which axis should be scaled.

    Returns:
        pd.DataFrame: The dataframe with scaled wavelength axis.
    """
    try:
        scaling = ureg(unit).to("nm").magnitude
        dataframe.index = dataframe.index.set_levels(
            dataframe.index.levels[1].astype(float) * scaling, level=1
        )

        return dataframe
    except UndefinedUnitError:
        logger.warning(
            "Unrecognized unit %s. "
            "The wavelength axis may not be scaled appropriately.",
            unit,
        )
        return dataframe
    except DimensionalityError:
        logger.warning(
            "Unit %s is not a valid wavelength unit. "
            "The wavelength axis may not be scaled appropriately.",
            unit,
        )
        return dataframe


def _read_wvase_dataframe(file_object: TextIO) -> pd.DataFrame:
    """
    Reads a wvase formated table from a TextIO file object.
    The pointer of the file object should point to the first line
    of the data table.

    Args:
        file_object (TextIO): The file object containing the wvase table.

    Returns:
        pd.DataFrame:
            Dataframe containing the psi/delta values.
            Ready to be further processed by pyElli.
    """
    dframe = pd.read_csv(
        file_object,
        sep="\t",
        header=None,
        names=["Wavelength", "Angle of Incidence", "Ψ", "Δ", "Ψ_err", "Δ_err"],
    )
    print(dframe)
    dframe = (
        dframe[dframe.apply(lambda x: is_float(x.iloc[0]), axis=1)]
        .set_index(["Wavelength", "Angle of Incidence"])
        .swaplevel(0, 1)
    )
    return dframe


def _read_complete_ease_dataframe(file_object: TextIO) -> pd.DataFrame:
    """
    Reads a complete ease formated table from a TextIO file object.
    The pointer of the file object should point to the first line
    of the data table.

    Args:
        file_object (TextIO): The file object containing the complete ease table.

    Returns:
        pd.DataFrame:
            Dataframe containing the psi/delta values.
            Ready to be further processed by pyElli.
    """
    dframe = pd.read_csv(file_object, sep="\t", header=None, index_col=(2, 1))
    dframe = dframe[dframe[0] == "E"]
    dframe = dframe.iloc[:, 1:]
    dframe.index.names = ("Angle of Incidence", "Wavelength")
    dframe.columns = ["Ψ", "Δ", "Ψ_err", "Δ_err"]
    return dframe


def read_woollam_psi_delta(fname: str) -> pd.DataFrame:
    r"""Read a psi/delta woollam ascii file.

    Args:
        fname (str): Filename of the measurement ascii file.

    Returns:
        pd.DataFrame: DataFrame containing the psi/delta data in
        the format to be further processes inside pyElli.
    """

    encoding = detect_encoding(fname)

    with open(fname, encoding=encoding) as fobj:
        line_number = fobj.tell()
        metadata = []
        file_format = ""
        line = fobj.readline()
        while line:
            if _is_wvase_tabular(line):
                line_number = fobj.tell()
                file_format = "wvase"
                break
            if _is_complete_ease_tabular(line):
                line_number = fobj.tell()
                file_format = "complete_ease"
                break
            if _is_tan_cos_format(line):
                raise NotImplementedError(
                    "The wvase Tan(Psi)/Cos(Delta) format is not supported. "
                    "Please try using wvase's psi/delta format."
                )
            metadata.append(line)
            line = fobj.readline()
        fobj.seek(line_number)

        if not line:
            raise ValueError(f"Invalid file format for {fname}")

        reader_map = {
            "wvase": _read_wvase_dataframe,
            "complete_ease": _read_complete_ease_dataframe,
        }

        if file_format not in reader_map:
            raise ValueError(f"Invalid file format {file_format} for file {fname}")
        data = reader_map.get(file_format)(fobj)

    data = scale_to_nm(metadata[-1], data)
    return data.iloc[:, :-2]


def read_woollam_rho(fname: str) -> pd.DataFrame:
    r"""Read a psi/delta woollam ascii file and converts it to rho values.

    Args:
        fname (str): Filename of the measurement ascii file.

    Returns:
        pd.DataFrame: DataFrame containing the rho data in
        the format to be further processes inside pyElli.
    """
    psi_delta = read_woollam_psi_delta(fname)
    return calc_rho(psi_delta)



# Encoding: utf-8
"""Dispersion law with gaussian oscillators."""

import numpy as np
import numpy.typing as npt
from numpy.lib.scimath import sqrt

# pylint: disable=no-name-in-module
from scipy.special import dawsn

from ..utils import conversion_wavelength_energy
from .base_dispersion import Dispersion


class Gaussian(Dispersion):
    r"""Dispersion law with gaussian oscillators.

    Single parameters:
        --

    Repeated parameters:
        :A: Amplitude of the oscillator. Defaults to 1.
        :E: Central energy. Defaults to 1. Unit in eV.
        :sigma: Broadening of the Gaussian. Defaults to 1. Unit in eV.

    Output:

        .. math::
            \varepsilon(E) = \sum_j & \; 2 \cdot \boldsymbol{A}_j / \sqrt{π} \cdot
                    (D\left(2 \cdot \sqrt{2 \cdot \ln(2)} \cdot (E + \boldsymbol{E}_j)
                    / \boldsymbol{sigma}_j\right) \\
                    &- D\left(2 \cdot \sqrt{2 \cdot \ln(2)} \cdot (E - \boldsymbol{E}_j)
                    / \boldsymbol{sigma}_j\right) \\
                    &+ i \cdot \Bigl(\boldsymbol{A}_j \cdot \exp\left(-(4 \cdot \ln(2) \cdot
                    (E - \boldsymbol{E}_j)/ \boldsymbol{sigma}_j\right)^2 \\
                    &- \boldsymbol{A}_j \cdot \exp\left(-(4 \cdot ln(2) \cdot
                    (E + \boldsymbol{E}_j)/ \boldsymbol{sigma}_j\right)^2\Bigr)

        D is the
        `Dawson function
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.dawsn.html>`_.
        The summation index :math:`j` is the index of the respective oscillator.

    References:
        * De Sousa Meneses, Malki, Echegut, J. Non-Cryst. Solids 351, 769-776 (2006)
        * Peiponen, Vartiainen, Phys. Rev. B. 44, 8301 (1991)
        * Fujiwara, Collins, Spectroscopic Ellipsometry for Photovoltaics Volume 1,
          Springer International Publishing AG, 2018, p. 137
    """

    single_params_template = {}
    rep_params_template = {"A": 1, "E": 1, "sigma": 1}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        energy = conversion_wavelength_energy(lbda)
        ftos = 2 * sqrt(np.log(2))
        return sum(
            2
            * c.get("A")
            / sqrt(np.pi)
            * (
                dawsn(ftos * (energy + c.get("E")) / c.get("sigma"))
                - dawsn(ftos * (energy - c.get("E")) / c.get("sigma"))
            )
            + 1j
            * (
                c.get("A")
                * np.exp(-((ftos * (energy - c.get("E")) / c.get("sigma")) ** 2))
                - c.get("A")
                * np.exp(-((ftos * (energy + c.get("E")) / c.get("sigma")) ** 2))
            )
            for c in self.rep_params
        )



import chardet


def detect_encoding(fname: str) -> str:
    r"""Detects the encoding of file fname.
    Args:
      fname (str): Filename
    Returns:
      str: Encoding identifier string.
    """
    with open(fname, "rb") as f:
        raw_data = f.read()
    result = chardet.detect(raw_data)
    return result["encoding"]



# Encoding: utf-8
import sys

from . import database as db
from .database.materials_db import AIR
from .dispersions import *
from .dispersions.base_dispersion import *
from .experiment import Experiment
from .importer.accurion import read_accurion_psi_delta
from .importer.nexus import *
from .importer.spectraray import *
from .importer.woollam import read_woollam_psi_delta, read_woollam_rho, scale_to_nm
from .materials import *
from .result import Result, ResultList
from .solver2x2 import Solver2x2
from .solver4x4 import *
from .structure import *
from .utils import *



# Encoding: utf-8
"""Polynomial dispersion."""

import numpy.typing as npt

from .base_dispersion import Dispersion


class Polynomial(Dispersion):
    r"""Polynomial expression for the dielectric function.

    Single parameters:
        :e0: Defaults to 1.

    Repeated parameters:
        :f: Defaults to 0.
        :e: Defaults to 0.

    Output:
        .. math::
            \varepsilon(\lambda) =
            \boldsymbol{\varepsilon_0} + \sum_j \boldsymbol{f}_j \cdot \lambda^{\boldsymbol{e}_j}
    """

    single_params_template = {"e0": 1}
    rep_params_template = {"f": 0, "e": 0}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return self.single_params.get("e0") + sum(
            c.get("f") * lbda ** c.get("e") for c in self.rep_params
        )



"""A helper class to load data from Accurion EP4 DAT files.
Typical files look like: Si3N4_on_4inBF33_W03_20240905-105631.ds.dat
"""

import numpy as np
import pandas as pd

from . import detect_encoding


def read_accurion_psi_delta(fname: str) -> pd.DataFrame:
    r"""Read a psi/delta Accurion dat file.

    Args:
      fname (str): Filename of the measured dat file

    Returns:
      pd.DataFrame: DataFrame containing the psi/delta data in the pyElli-compatible format.
    """
    encoding = detect_encoding(fname)
    psi_delta_df = pd.read_csv(
        fname, delimiter="\t", encoding=encoding, skiprows=0, header=0
    )[1:].astype(float)
    psi_delta_df = psi_delta_df.reindex(columns=list(["AOI", "Lambda", "Delta", "Psi"]))
    psi_delta_df = psi_delta_df.rename(
        columns={
            "AOI": "Angle of Incidence",
            "Lambda": "Wavelength",
            "Delta": "Δ",
            "Psi": "Ψ",
        }
    )
    psi_delta_df = psi_delta_df.groupby(["Angle of Incidence", "Wavelength"]).sum()

    # wrap delta range
    psi_delta_df.loc[:, "Δ"] = psi_delta_df.loc[:, "Δ"].where(
        psi_delta_df.loc[:, "Δ"] <= 180, psi_delta_df.loc[:, "Δ"] - 360
    )

    return psi_delta_df



#!/usr/bin/python
# encoding: utf-8

import elli
import numpy as np
from numpy.lib.scimath import sqrt
from scipy.constants import pi
from scipy.signal import argrelmax, argrelmin


def test_cholesteric_lc():
    # Materials
    front = back = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=1.6))

    # Liquid crystal oriented along the x direction
    (no, ne) = (1.5, 1.7)
    Dn = ne - no
    n_med = (ne + no) / 2
    LC = elli.UniaxialMaterial(
        elli.ConstantRefractiveIndex(n=no), elli.ConstantRefractiveIndex(n=ne)
    )  # ne along z
    R = elli.rotation_v_theta(elli.E_Y, 90)  # rotation round y
    LC.set_rotation(R)  # apply rotation from z to x
    # Cholesteric pitch:
    p = 650
    # One half turn of a right-handed helix:
    TN = elli.TwistedLayer(LC, p / 2, 25, 180)

    # Inhomogeneous layer, repeated layer, and structure
    N = 5  # number half pitch repetitions
    h = N * p / 2
    L = elli.RepeatedLayers([TN], N)
    s = elli.Structure(front, [L], back)

    # Normal incidence:
    Kx = 0.0

    # Calculation parameters
    lbda_min, lbda_max = 600, 1500  # (nm)
    lbda = np.linspace(lbda_min, lbda_max, 100)
    k0 = 2 * pi / (lbda * 1e-9)

    # Theoretical calculation
    q = 2 * pi / p / 1e-9
    alpha = q / k0
    epsilon = (no**2 + ne**2) / 2
    delta = (no**2 - ne**2) / 2
    n2 = sqrt((alpha**2 + epsilon - sqrt(4 * epsilon * alpha**2 + delta**2)))
    w = 1j * (ne**2 - n2**2 - alpha**2) / (2 * alpha * n2)  # not k0/c
    A = -2j * k0 * n2 * h * 1e-9

    R_th = (
        np.abs(
            (w**2 + 1)
            * (1 - np.exp(-2j * k0 * n2 * h * 1e-9))
            / (
                2 * w * (1 + np.exp(-2j * k0 * n2 * h * 1e-9))
                - 1j * (w**2 - 1) * (1 - np.exp(-2j * k0 * n2 * h * 1e-9))
            )
        )
        ** 2
    )

    # Berreman simulation
    data = s.evaluate(lbda, 0)
    R_RR = data.Rc_RR

    # Checks positions of local extrema
    np.testing.assert_allclose(argrelmax(R_RR)[0], argrelmax(R_th)[0], atol=3)
    np.testing.assert_allclose(argrelmin(R_RR)[0], argrelmin(R_th)[0])
    np.testing.assert_array_almost_equal(R_RR, R_th, decimal=1)


def test_twisted_nematic_lc():
    # Materials
    glass = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=1.55))
    front = back = glass

    # Liquid crystal oriented along the x direction
    (no, ne) = (1.5, 1.6)
    Dn = ne - no
    LC = elli.UniaxialMaterial(
        elli.ConstantRefractiveIndex(n=no), elli.ConstantRefractiveIndex(n=ne)
    )
    R = elli.rotation_v_theta(elli.E_Y, 90)
    LC.set_rotation(R)
    d = 4330
    TN = elli.TwistedLayer(LC, d, 18, 90)

    # Structure
    s = elli.Structure(front, [TN], back)

    # Calculation parameters
    (lbda_min, lbda_max) = (200e-9, 1)  # (m)
    k0_list = np.linspace(2 * pi / lbda_max, 2 * pi / lbda_min)
    lbda_list = (2 * pi) / k0_list * 1e9

    # Gooch-Tarry law
    u = 2 * d * Dn / lbda_list
    T_gt = np.sin(pi / 2 * sqrt(1 + u**2)) ** 2 / (1 + u**2)

    # Berreman simulation
    data = s.evaluate(lbda_list, 0)
    T_bm = data.T_pp

    # Compare results
    np.testing.assert_array_almost_equal(T_bm, T_gt, decimal=2)



"""
Basic usage
===========

Basic usage of building a model and fitting it to measurement data of SiO2 on Si.

"""

# %%
import elli
from elli.fitting import ParamsHist, fit

# sphinx_gallery_thumbnail_path = '_static/basic_usage.png'


# %%
# Reading data
# ------------------------
#
# We load the data from the generated
# `NeXus file <https://manual.nexusformat.org/classes/contributed_definitions/NXellipsometry.html#nxellipsometry>`_
# and select the angle we want to analyse.
# You may set the ANGLE constant to 50 or 60 to select
# other angles of incidence from the example file.
# Additionally, we're cutting the wavelength axis to be in between 210 nm and 800 nm.
# This is because we're using literature values for Si,
# which are only defined in this wavelength range.
ANGLE = 70
psi_delta = elli.read_nexus_psi_delta("SiO2onSi.ellips.nxs").loc[ANGLE][210:800]

# %%
# Setting parameters
# ------------------------
#
# As an example we analyse an oxidation layer of SiO2 on Si.
# Prior to defining our model, we have to set the parameters we want to use.
# We're going to use a :ref:`Cauchy model <cauchy>` for SiO2 and load the Si values from
# `literature values <https://refractiveindex.info/?shelf=main&book=Si&page=Aspnes>`_.
# The parameter names can be chosen freely,
# but you have to use the exact same name in the later model definition.
# The package uses lmfit as fitting tool and you may refer to their
# `documentation <https://lmfit.github.io/lmfit-py/parameters.html#lmfit.parameter.Parameters.add>`_
# for details on parameter definition.

params = ParamsHist()
params.add("SiO2_n0", value=1.452, min=-100, max=100, vary=False)
params.add("SiO2_n1", value=36.0, min=-40000, max=40000, vary=False)
params.add("SiO2_n2", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_k0", value=0, min=-100, max=100, vary=False)
params.add("SiO2_k1", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_k2", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_d", value=20, min=0, max=40000, vary=True)

# %%
# Load silicon dispersion from the refractiveindexinfo database
# -------------------------------------------------------------
# You can load any material from the index
# `refractiveindex.info <https://refractiveindex.info>`__, which is
# embedded into the software (so you may use it offline, too). Here, we
# are interested in the literature values for the silicon substrate.
# First we need to load the database with ``rii_db = elli.db.RII()`` and
# then we can query it with ``rii_db.get_mat("Si", "Aspnes")`` to load
# this
# `entry <https://refractiveindex.info/?shelf=main&book=Si&page=Aspnes>`__.
rii_db = elli.db.RII()
Si = rii_db.get_mat("Si", "Aspnes")


# %%
# Building the model
# ------------------------
#
# For simple parameter estimation,
# the fit decorator (**@fit**) in conjunction with the model definition is used.
# The fitting decorator takes a pandas dataframe containing
# the psi/delta measurement data (**psi_delta**) and the model parameters (**params**) as an input.
# It then passes the wavelength from measurement dataframe (**lbda**)
# and the parameters to the actual model function.
#
# Inside the model function the optical model is built,
# i.e. the Si literature values are loaded
# and the fitting parameters are filled into the Cauchy dispersion.
# For details on how to insert data into the Cauchy model or other optical dispersion models,
# you may refer to the documentation of pyElli.
# Please keep in mind that the parameters you use here
# have to be defined in the parameter object **param**.
#
# From the dispersion model isotropic materials are generated
# (could also be an anisotropic material, refer to the docs for an overview).
# This is done by calling the :code:`elli.IsotropicMaterial(...)` function
# with a dispersion model as a parameter
# or simply calling :code:`.get_mat()` on a dispersion model.
# These two approaches are equivalent. From these materials the layer is build,
# which only consists of the SiO2 layer in this example.
# The final structure consists of an incoming half-space,
# the layers and an outgoing half space. Specifically,
# typically the light is coming from air and finally gets absorbed by the bulk material,
# in our example this is Si, i.e. we call :code:`elli.Structure(elli.AIR, Layer, Si)`.
#
# To provide simulated data, we have to evaluate the structure
# by calling the :code:`evaluate(...)` function,
# which takes the experimental wavelength array **lbda**, **ANGLE** under which the experiment
# was performed and the solver to be used to solve the transfer-matrix problem.
# Here, we use a simple 2x2 matrix approach,
# which splits the interaction in s and p-parts and therefore cannot account for anisotropy.
# There exist 4x4 matrix solvers as well.
# You may refer to the :ref:`solver documentation <solvers>` for further details.
#
# Executing the cell below in a jupyter notebook displays a comparison of the simulated Ψ / Δ values
# at the current parameter values with their measured counterparts.
# Additionally, input fields for each model parameter are shown.
# You may change the parameters and the calculated data will change accordingly.
# For clarification the modeled data is shown with `_calc` postfix in the legend.
@fit(psi_delta, params)
def model(lbda, params):
    # Generate the cauchy model from the current lmfit parameters
    SiO2 = elli.Cauchy(
        params["SiO2_n0"],
        params["SiO2_n1"],
        params["SiO2_n2"],
        params["SiO2_k0"],
        params["SiO2_k1"],
        params["SiO2_k2"],
    ).get_mat()
    # get_mat() generates an IsotropicMaterial from the dispersion relation

    # Construct the layers you expect in your sample
    # Here, it only consists of one layer SiO2 in between air and Si.
    # We build the structure coming from air, through the layers,
    # represented as an array, and having Si as bulk material.
    structure = elli.Structure(
        elli.AIR,  # Input medium
        [elli.Layer(SiO2, params["SiO2_d"])],  # Overlayer structure
        Si,
    )  # Output medium / Substrate

    # The model should return the evaluation of the structure at the experimental wavelengths lbda,
    # the experimental angle ANGLE and it should define a solver to calculate the transfer matrix.
    return structure.evaluate(lbda, ANGLE, solver=elli.Solver2x2)


# %%
# Fit and plot fit results
# ------------------------
#
# The fit of the data can be executed by calling the fit() function on the model function,
# which automatically gets attached by the @fit decorator in the cell above.
# The following cell basically executes the fit and plots
# a comparison between the measurement and fitted data.
fit_stats = model.fit()
model.plot()

# %%
# Extracting the optical properties from the fit
# ----------------------------------------------
#
# Since we want to extract the dispersion relation of a layer in our measured stack,
# we can use our fitted parameters.
# The fit parameters are contained in the fits output :code:`params` attribute,
# i.e. :code:`fit_stats.params` for our example.
# We can use it to call our dispersion relation we used in our model
# (here it is a Cauchy dispersion relation)
# and fill in our fitted value.
# By calling :code:`get_dielectric_df()` we can get the dielectric function of the material,
# which is plotted here as an example. :code:`get_dielectric_df` uses a default wavelength range
# which can also be changed by inputting a wavelength array as a parameter.
fitted_model = elli.Cauchy(
    fit_stats.params["SiO2_n0"],
    fit_stats.params["SiO2_n1"],
    fit_stats.params["SiO2_n2"],
    fit_stats.params["SiO2_k0"],
    fit_stats.params["SiO2_k1"],
    fit_stats.params["SiO2_k2"],
)

fitted_model.get_dielectric_df().plot(backend="plotly")

# %%
# We can also call :code:`get_refractive_index_df()`
# to get the refractive index of the material as dataframe.
fitted_model.get_refractive_index_df().plot(backend="plotly")

# %%
# If you want to write your data to a file you simply call pandas :code:`to_csv(...)`
# function to write a csv file, i.e. for the dielectric function this writes as
fitted_model.get_dielectric_df().to_csv("SiO2_diel_func.csv")

# %%
# You may also access a single value of your optical model
fit_stats.params["SiO2_n0"].value

# %%
# Our simply print the fitted values in a list together with their fitting errors
fit_stats.params

# %%
# Show fit statistics
# --------------------
# Now, we may also print out the fit statictics from the model fit in the cell above.
# The fit statistics are simple `lmfit fit statistics <https://lmfit.github.io/lmfit-py/fitting.html#>`_, too.
# Typically, one uses chi square values as a figure of merit for the fit quality.
# It is stored in the chisqr attribute of the fit_stats variable we defined above.
fit_stats.chisqr

# %%
# We can print the full fit statistics, too.
fit_stats

# %%
# References
# ----------
# `Here <https://github.com/PyEllips/pyElli/tree/master/examples/Basic%20Usage>`_
# you can find the latest jupyter notebook and data files of this example.



"""Test Kramers Kronig relations"""

import elli
import numpy as np
from elli.kkr import im2re, im2re_reciprocal
from numpy.testing import assert_array_almost_equal


def test_tauc_lorentz():
    """Test whether the kkr reproduces the analytical expression of Tauc-Lorentz"""
    lbda = np.linspace(1e-2, 2000, 2000)
    g = elli.TaucLorentz(Eg=5).add(A=20, E=8, C=5)
    assert_array_almost_equal(
        im2re_reciprocal(g.get_dielectric(lbda).imag, lbda),
        g.get_dielectric(lbda).real,
        decimal=6,
    )


def test_tauc_lorentz_energy():
    """Test whether the kkr in non reciprocal formulation reproduces the analytical expression
    of Tauc-Lorentz"""
    energy = np.linspace(0, 10, 5000)
    amp = 20
    osc_energy = 5
    gamma = 0.1
    lorentz = amp / (osc_energy**2 - energy**2 - 1j * gamma * energy)

    assert_array_almost_equal(
        im2re(lorentz.imag, energy)[10:], lorentz.real[10:], decimal=2
    )


def test_lorentz():
    """Test whether the kkr reproduces the analytical expression of a Lorentz oscillator"""
    lbda = np.linspace(1e-2, 5000, 5000)
    g = elli.LorentzEnergy().add(A=20, E=5, gamma=5)
    assert_array_almost_equal(
        1 + im2re_reciprocal(g.get_dielectric(lbda).imag, lbda)[:-1000],
        g.get_dielectric(lbda).real[:-1000],
        decimal=2,
    )


def test_gauss():
    """KKR reproduces the analytical expression of gaussian."""
    lbda = np.linspace(1e-2, 5000, 5000)
    g = elli.Gaussian().add(A=10, E=8, sigma=5)
    assert_array_almost_equal(
        im2re_reciprocal(g.get_dielectric(lbda).imag, lbda)[:-1000],
        g.get_dielectric(lbda).real[:-1000],
        decimal=2,
    )



"""Abstract base class for Decorator functions for convenient fitting"""

# Encoding: utf-8
from abc import ABC, abstractmethod

import pandas as pd

try:
    from lmfit import Parameters
    from ipywidgets import widgets
except ImportError as e:
    raise ImportError(
        "This module requires lmfit and ipywidgets to work properly.\n"
        "Try installing this package with the additional fitting requirement, "
        "i.e. pip install pyElli[fitting]"
    ) from e

from .params_hist import ParamsHist


def is_in_notebook() -> bool:
    """Checks whether the current shell is in a jupyter notebook.

    Returns:
        bool: True if the shell is in jupyter, False otherwise.
    """
    try:
        # pylint: disable=import-outside-toplevel
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


class FitDecorator(ABC):
    """The abstract base class for fitting decorators.
    Providing features for fit, undo and redo buttons."""

    def __init__(self) -> None:
        self.params = ParamsHist()
        self.fitted_params = ParamsHist()
        self.last_params = ParamsHist()
        self.initial_params = ParamsHist()
        self.param_widgets = {}
        self.fit_kwargs = {}

    @abstractmethod
    def get_model_data(
        self, params: Parameters = None, append_exp_data=False
    ) -> pd.DataFrame:
        """Gets the data from the provided model with the provided parameters.
        If no parameters are provided, the fitted parameters are used
        (which default to the initial parameters if no fit has been triggered).

        Args:
            params (Parameters, optional):
                The parameters to calculate the model with.
                If not provided, the fitted parameters are used.
                Defaults to None.
            append_exp_data (bool, optional):
                Appends the experimental data if set to True.
                Defaults to False.

        Returns:
            pd.DataFrame: The model results
        """

    def to_csv(
        self,
        *args,
        fname: str,
        params: Parameters = None,
        append_exp_data: bool = False,
        **kwargs,
    ) -> None:
        """Saves the current model to csv. This is just a wrapper to
        pandas Dataframe and any argument to pandas to_csv may be passed
        as function arguments.

        Args:
            fname (str): The file name to save the data to.
            params (Parameters, optional):
                The parameters to calculate the model with.
                If not provided, the fitted parameters are used.
                Defaults to None.
            append_exp_data (bool, optional):
                Appends the experimental data if set to True.
                Defaults to False.
        """
        self.get_model_data(params, append_exp_data).to_csv(fname, *args, **kwargs)

    @abstractmethod
    def fit(self, method: str = "") -> None:
        """Execute lmfit with the current fitting parameters

        Args:
            method (str, optional): The fitting method to use.
                                    Any method supported by scipys curve_fit is allowed.
                                    Defaults to 'leastsq'.

        Returns:
            Result: The fitting result
        """

    @abstractmethod
    def update_selection(self, change: dict = None) -> None:
        """Update plot after selection of displayed data

        Args:
            change (dict, optional): A dictionary containing the ipywidgets change event
        """

    def set_vary_param(self, change: dict) -> None:
        self.initial_params[change.owner.description_tooltip].vary = change.new
        self.params[change.owner.description_tooltip].vary = change.new

    def fit_button_clicked(self) -> None:
        """Fit and update plot after the fit button has been clicked

        Args:
            selected (dict): Dict containing the current widget information
                of the selection dropdown.
        """
        self.fit(**self.fit_kwargs)
        if isinstance(self.params, ParamsHist):
            self.params.commit()
        self.params.update_params(self.fitted_params)
        self.update_widgets()

        self.update_selection()

    def re_undo_button_clicked(self, button: widgets.Button) -> None:
        """Redo or undo an operation on the parameters history object

        Args:
            button (widgets.Button):
                The button instance, which triggered the event.
        """
        if not isinstance(self.params, ParamsHist):
            return

        if button.description == "Undo":
            self.last_params = self.params.pop()
            self.update_widgets()
        elif self.last_params is not None:
            self.params.update_params(self.last_params)
            self.update_widgets()
            self.last_params = None

        self.update_selection()

    def reset_to_init_params(self) -> None:
        """Resets the parameters to the initial values"""

        if not isinstance(self.params, ParamsHist):
            return

        self.last_params = self.params.pop()
        self.params = self.initial_params.copy()
        self.update_widgets()
        self.update_selection()

    def update_params(self, change: dict) -> None:
        """Update plot after a change of fitting parameters

        Args:
            change (dict): A dictionary containing the ipywidgets change event
            selected (dict): The selected value of the data display dropdown widget
        """
        if isinstance(self.params, ParamsHist):
            self.params.commit()
        self.params[change.owner.description].value = change.new

        self.update_selection()

    def update_widgets(self) -> None:
        """Updates the widget values according to the current parameters."""
        for param in self.params:
            widget = self.param_widgets.get(param)
            if widget is not None:
                # Unobserve and re-observe, otherwise the update_params
                # would be called on every change
                # and a history event would be triggered for each of these changes.
                widget.unobserve(self.update_params, names="value")
                widget.value = self.params[param].value
                widget.observe(self.update_params, names="value")



"""
Custom fitting example
======================

This is a short example for fitting data, which is not covered by the fitting widget.
It relies on calling lmfit manually. Therefore one needs to provide a fitting function, which calculates the numerical residual between measurement and model.
This notebook is about fitting multiple datasets from different angles of incidents simultaneously, but can be adapted to a wide range of different use cases.

"""

# %%
import numpy as np
import elli
from elli.fitting import ParamsHist
from lmfit import minimize, fit_report
import matplotlib.pyplot as plt

# sphinx_gallery_thumbnail_path = '_static/custom_fitting.png'


# %%
# Data import
# -----------
data = elli.read_nexus_psi_delta("SiO2onSi.ellips.nxs").loc[
    (slice(None), slice(210, 800)), :
]
lbda = data.loc[50].index.get_level_values("Wavelength").to_numpy()
data

# %%
# Setting up invariant materials and fitting parameters
# -----------------------------------------------------
rii_db = elli.db.RII()
Si = rii_db.get_mat("Si", "Aspnes")

params = ParamsHist()
params.add("SiO2_n0", value=1.452, min=-100, max=100, vary=True)
params.add("SiO2_n1", value=36.0, min=-40000, max=40000, vary=True)
params.add("SiO2_n2", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_k0", value=0, min=-100, max=100, vary=False)
params.add("SiO2_k1", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_k2", value=0, min=-40000, max=40000, vary=False)
params.add("SiO2_d", value=20, min=0, max=40000, vary=True)


# %%
# Model helper function
# ---------------------
# This model function is not strictly needed, but simplifies the fit function, as the model only needs to be defined once.
def model(lbda, angle, params):
    SiO2 = elli.Cauchy(
        params["SiO2_n0"],
        params["SiO2_n1"],
        params["SiO2_n2"],
        params["SiO2_k0"],
        params["SiO2_k1"],
        params["SiO2_k2"],
    ).get_mat()

    structure = elli.Structure(
        elli.AIR,
        [elli.Layer(SiO2, params["SiO2_d"])],
        Si,
    )

    return structure.evaluate(lbda, angle, solver=elli.Solver2x2)


# %%
# Defining the fit function
# -------------------------
# The fit function follows the protocol defined by the lmfit package and needs the parameters dictionary as first argument.
# It has to return a residual value, which will be minimized. Here psi and delta are used to calculate the residual, but could be changed to transmission or reflection data.


def fit_function(params, lbda, data):
    residual = []

    for phi_i in [50, 60, 70]:
        model_result = model(lbda, phi_i, params)

        resid_psi = data.loc[(phi_i, "Ψ")].to_numpy() - model_result.psi
        resid_delta = data.loc[(phi_i, "Δ")].to_numpy() - model_result.delta

        residual.append(resid_psi)
        residual.append(resid_delta)

    return np.concatenate(residual)


# %%
# Running the fit
# ---------------
# The fitting is performed by calling the minimize function with the fit_function and the needed arguments.
# It is possible to change the underlying algorithm by providing the method kwarg.

out = minimize(fit_function, params, args=(lbda, data), method="leastsq")
print(fit_report(out))

# %%
# Plotting the results
# ----------------------------------------------

fit_50 = model(lbda, 50, out.params)
fit_60 = model(lbda, 60, out.params)
fit_70 = model(lbda, 70, out.params)

fig = plt.figure(dpi=100)
ax = fig.add_subplot(1, 1, 1)
ax.scatter(lbda, data.loc[(50, "Ψ")], s=20, alpha=0.1, label="50° Measurement")
ax.scatter(lbda, data.loc[(60, "Ψ")], s=20, alpha=0.1, label="Psi 60° Measurement")
ax.scatter(lbda, data.loc[(70, "Ψ")], s=20, alpha=0.1, label="Psi 70° Measurement")
(psi50,) = ax.plot(lbda, fit_50.psi, c="tab:blue", label="Psi 50°")
(psi60,) = ax.plot(lbda, fit_60.psi, c="tab:orange", label="Psi 60°")
(psi70,) = ax.plot(lbda, fit_70.psi, c="tab:green", label="Psi 70°")
ax.set_xlabel("wavelenth / nm")
ax.set_ylabel("psi / degree")
ax.legend(handles=[psi50, psi60, psi70], loc="lower left")
fig.canvas.draw()

fig = plt.figure(dpi=100)
ax = fig.add_subplot(1, 1, 1)
ax.scatter(lbda, data.loc[(50, "Δ")], s=20, alpha=0.1, label="Delta 50° Measurement")
ax.scatter(lbda, data.loc[(60, "Δ")], s=20, alpha=0.1, label="Delta 60° Measurement")
ax.scatter(lbda, data.loc[(70, "Δ")], s=20, alpha=0.1, label="Delta 70° Measurement")
(delta50,) = ax.plot(lbda, fit_50.delta, c="tab:blue", label="Delta 50°")
(delta60,) = ax.plot(lbda, fit_60.delta, c="tab:orange", label="Delta 60°")
(delta70,) = ax.plot(lbda, fit_70.delta, c="tab:green", label="Delta 70°")
ax.set_xlabel("wavelenth / nm")
ax.set_ylabel("delta / degree")
ax.legend(handles=[delta50, delta60, delta70], loc="lower right")
fig.canvas.draw()



"""This modules creates a formula parser"""

from functools import lru_cache
import os
from operator import add, mul, neg, sub, truediv
from typing import Dict

import numpy as np
import scipy.constants as sc
from lark import Lark, Transformer, v_args
from scipy.special import dawsn  # pylint: disable=no-name-in-module


@v_args(inline=True)
class FormulaTransformer(Transformer):
    """Transformer class for parsing formulas"""

    single_params: Dict[str, float]
    repeated_params: Dict[str, np.ndarray]
    no_repeated_params: int

    number = float
    add = add
    sub = sub
    mul = mul
    div = truediv
    neg = neg
    power = pow

    def _check_and_set(self, repeated_params):
        if not repeated_params:
            self.no_repeated_params = 0
            return

        if not isinstance(repeated_params, dict):
            raise ValueError(
                f"Repeated parameters must be a dict but found {type(repeated_params)}"
            )

        params = iter(repeated_params.items())
        length = len(next(params)[1])
        for name, param in params:
            if not isinstance(param, np.ndarray):
                raise TypeError(
                    f"Expected {name} to be of type numpy ndarray but found type {type(param)}"
                )
            if length != len(param):
                raise ValueError("Repeated parameters must have all the same length.")

        self.no_repeated_params = length

    def _check_and_set_single(self, single_params):
        for name, param in single_params.items():
            if not isinstance(param, (float, int)):
                raise TypeError(
                    f"Expected {name} to be of type float but found type {type(param)}."
                )
        self.single_params = single_params

    def __init__(
        self,
        x_axis_name: str,
        x_axis_values: np.ndarray,
        single_params: Dict[str, float],
        repeated_params: Dict[str, np.ndarray],
    ):
        super().__init__()
        if not isinstance(x_axis_name, str):
            raise TypeError("x_axis_name must be a string.")

        if not isinstance(x_axis_values, np.ndarray):
            raise TypeError("x_axis_values must be a numpy array.")

        self.x_axis_name = x_axis_name
        self.x_axis_values = x_axis_values

        self._check_and_set(repeated_params)
        self._check_and_set_single(single_params)

        self.repeated_params = repeated_params
        self.single_params = single_params

    def eps(self, inp):
        """Return an epsilon type formula"""
        return "eps", inp

    # pylint: disable=invalid-name
    def n(self, inp):
        """Return an index type formula"""
        return "n", inp

    def kkr_term(self, term):
        """Calculate the kramers kronig transformation on the function"""
        raise NotImplementedError("kkr transformation not yet implemented")

    def func(self, name, val):
        """Evaluates a function"""
        names = {
            "sin": np.sin,
            "cos": np.cos,
            "tan": np.tan,
            "sqrt": np.emath.sqrt,
            "dawsn": dawsn,
            "ln": np.log,
            "log": np.log10,
            "heaviside": np.heaviside,
        }

        if name in names:
            return names[name](val)

        raise ValueError(f"Unknown function: {name}")

    def builtin(self, name):
        """Returns the values for builtin tokens"""
        names = {
            "1j": 1j,
            "pi": sc.pi,
            "eps_0": sc.epsilon_0,
            "hbar": sc.hbar,
            "h": sc.h,
            "c": sc.c,
        }

        if name in names:
            return names[name]

        raise ValueError(f"Unknown constant: {name}")

    def sum_expr(self, expr):
        """Sum an expression"""
        return expr.sum(axis=1)

    def single_param_name(self, name):
        """Return a parameter inside a non-repeated section"""
        if name == self.x_axis_name:
            return self.x_axis_values

        if name in self.single_params:
            return self.single_params[name]

        raise ValueError(f"No such parameter {name}")

    def param_name(self, name):
        """Return a parameter inside a repeated section"""
        if name == self.x_axis_name:
            return np.einsum(
                "i,j->ij", self.x_axis_values, np.ones(self.no_repeated_params)
            )

        if name in self.single_params:
            return self.single_params[name]

        if name in self.repeated_params:
            return self.repeated_params[name]

        raise ValueError(f"No such parameter {name}")


__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

with open(
    f"{__location__}/dispersion_function_grammar.lark", encoding="utf-8"
) as formula_grammar:
    grammar = Lark(
        formula_grammar,
        start="assignment",
        parser="lalr",
    )


@lru_cache(maxsize=128)
def parse_formula(formula: str):
    """
    Parses a dispersion formula string into an abstract syntax tree.
    Uses caching to avoid re-parsing of the formula.

    Args:
        formula (str): The formula string to parse.

    Returns:
        Lark.Tree: The parsed formula Tree.
    """
    return grammar.parse(formula)



"""Tests for reading accurion files"""

import pytest
from fixtures import datadir  # pylint: disable=unused-import

import elli


# pylint: disable=redefined-outer-name
def test_reading_of_psi_delta_woollam(datadir):
    """Psi/delta Accurion file is read w/o errors"""
    data = elli.read_accurion_psi_delta(
        datadir / "Si3N4_on_4inBF33_W02_20240903-150451.ds.dat"
    )

    assert data.shape == (114, 2)



# Encoding: utf-8
"""Cauchy dispersion with custom exponents."""

import numpy.typing as npt

from .base_dispersion import IndexDispersion


class CauchyCustomExponent(IndexDispersion):
    r"""Cauchy dispersion with custom exponents.

    Single parameters:
        :n0: Defaults to 1.5.

    Repeated parameters:
        :f: Defaults to 0.
        :e: Defaults to 1.

    Output:
        .. math::
            \varepsilon^{1/2}(\lambda) =
            \boldsymbol{n_0} + \sum_j \boldsymbol{f}_j \cdot \lambda^{\boldsymbol{e}_j}
    """

    single_params_template = {"n0": 1.5}
    rep_params_template = {"f": 0, "e": 1}

    def refractive_index(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return self.single_params.get("n0") + sum(
            c.get("f") * lbda ** c.get("e") for c in self.rep_params
        )



"""Decorator functions for convenient fitting"""

# Encoding: utf-8
from sys import float_info
from typing import Callable

import numpy as np
import numpy.typing as npt
import pandas as pd

try:
    from lmfit import Parameters, minimize
    import plotly.graph_objects as go
    from IPython.display import display
    from ipywidgets import widgets
except ImportError as e:
    raise ImportError(
        "This module requires lmfit, plotly, ipywidgets and ipython to be installed.\n"
        "Try installing this package with the additional fitting requirement, "
        "i.e. pip install pyElli[fitting]"
    ) from e

from ..result import Result
from ..utils import calc_pseudo_diel, calc_rho
from .decorator import FitDecorator, is_in_notebook
from .params_hist import ParamsHist


class FitRho(FitDecorator):
    """
    A class to fit psi/delta or rho based ellipsometry data with two degress of freedom
    """

    def set_psi_delta(
        self, update_exp: bool = False, update_names: bool = False
    ) -> None:
        """Sets Plot to Psi/Delta values

        Args:
            update_exp (bool, optional):
                Flag to change the experimental data as well. Defaults to False.
            update_names (bool, optional):
                Flag to change the label names. Defaults to False.
        """
        data = self.model(self.exp_data.index, self.params)
        self.fig.update_layout(yaxis_title="Ψ/Δ (°)")
        self.fig.data[2].y = data.psi
        self.fig.data[3].y = data.delta

        if update_exp:
            self.fig.data[0].y = self.exp_data.loc[:, "Ψ"]
            self.fig.data[1].y = self.exp_data.loc[:, "Δ"]

        if update_names:
            self.fig.data[0].name = "Ψ"
            self.fig.data[1].name = "Δ"
            self.fig.data[2].name = "Ψ_calc"
            self.fig.data[3].name = "Δ_calc"

    def set_rho(self, update_exp: bool = False, update_names: bool = False) -> None:
        """Sets Plot to Rho values

        Args:
            update_exp (bool, optional):
                Flag to change the experimental data as well. Defaults to False.
            update_names (bool, optional):
                Flag to change the label names. Defaults to False.
        """
        data = self.model(self.exp_data.index, self.params)
        self.fig.update_layout(yaxis_title="ρ")
        self.fig.data[2].y = data.rho.real
        self.fig.data[3].y = data.rho.imag

        if update_exp:
            exp_rho = calc_rho(self.exp_data)
            self.fig.data[0].y = exp_rho.apply(lambda x: x.real).values
            self.fig.data[1].y = exp_rho.apply(lambda x: x.imag).values

        if update_names:
            self.fig.data[0].name = "ρr"
            self.fig.data[1].name = "ρi"
            self.fig.data[2].name = "ρr_calc"
            self.fig.data[3].name = "ρi_calc"

    def set_residual(
        self,
        update_exp: bool = False,  # pylint: disable=unused-argument
        update_names: bool = False,
    ) -> None:
        """Sets plots to residual values

        Args:
            update_exp (bool, optional):
                Flag to change the experimental data as well. Defaults to False.
            update_names (bool, optional):
                Flag to change the label names. Defaults to False.
        """
        data = self.model(self.exp_data.index, self.params)
        self.fig.update_layout(yaxis_title="Residual")

        exp_rho = calc_rho(self.exp_data)
        self.fig.data[0].y = self.exp_data.loc[:, "Ψ"] - data.psi
        self.fig.data[1].y = self.exp_data.loc[:, "Δ"] - data.delta
        self.fig.data[2].y = exp_rho.apply(lambda x: x.real).values - data.rho.real
        self.fig.data[3].y = exp_rho.apply(lambda x: x.imag).values - data.rho.imag

        if update_names:
            self.fig.data[0].name = "Psi Res."
            self.fig.data[1].name = "Delta Res."
            self.fig.data[2].name = "ρr Res."
            self.fig.data[3].name = "ρi Res."

    def set_pseudo_diel(
        self, update_exp: bool = False, update_names: bool = False
    ) -> None:
        """Sets Plot to Pseudo Dielectric function values

        Args:
            update_exp (bool, optional): Flag to change the experimental data as well.
                                         Defaults to False.
            update_names (bool, optional): Flag to change the label names.
                                           Defaults to False.
        """
        data = self.model(self.exp_data.index, self.params)
        peps = calc_pseudo_diel(
            pd.DataFrame(data.rho, index=self.exp_data.index).iloc[:, 0], self.angle
        )
        self.fig.update_layout(yaxis_title="ϵ")
        self.fig.data[2].y = peps.loc[:, "ϵ1"]
        self.fig.data[3].y = peps.loc[:, "ϵ2"]

        if update_exp:
            exp_peps = calc_pseudo_diel(calc_rho(self.exp_data), self.angle)
            self.fig.data[0].y = exp_peps.loc[:, "ϵ1"]
            self.fig.data[1].y = exp_peps.loc[:, "ϵ2"]

        if update_names:
            self.fig.data[0].name = "ϵ1"
            self.fig.data[1].name = "ϵ2"
            self.fig.data[2].name = "ϵ1_calc"
            self.fig.data[3].name = "ϵ2_calc"

    def update_selection(self, change=None) -> None:
        """Update plot after selection of displayed data

        Args:
            change (dict, optional): A dictionary containing the ipywidgets change event
        """
        with self.fig.batch_update():
            update = self.update_dict.get(
                change.new if change is not None else self.selector.value
            )
            if update is not None:
                update(update_exp=True, update_names=True)

    def create_widgets(self) -> None:
        """Create ipywidgets for parameter estimation"""
        self.param_widgets = {}
        self.selector = widgets.Dropdown(
            options=["Psi/Delta", "Rho", "Pseudo Diel.", "Residual"],
            value="Psi/Delta",
            description="Display: ",
            disabled=False,
        )
        self.selector.observe(self.update_selection, names="value")

        checkboxes = []
        for param in self.params.valuesdict():
            curr_checkbox = widgets.Checkbox(
                value=self.params[param].vary,
                description_tooltip=param,
                layout=widgets.Layout(width="15px"),
                indent=False,
            )
            curr_checkbox.observe(self.set_vary_param, names="value")
            curr_widget = widgets.BoundedFloatText(
                self.params[param],
                min=(
                    float_info.min
                    if self.params[param].min == -float("inf")
                    else self.params[param].min
                ),
                max=(
                    float_info.max
                    if self.params[param].max == float("inf")
                    else self.params[param].max
                ),
                description=param,
                continuous_update=False,
            )
            curr_widget.observe(self.update_params, names="value")
            checkboxes.append(curr_checkbox)
            self.param_widgets[param] = curr_widget

        fit_button = widgets.Button(description="Fit")
        fit_button.on_click(lambda _: self.fit_button_clicked())
        button_list = [self.selector, fit_button]

        if isinstance(self.params, ParamsHist):
            undo_button = widgets.Button(description="Undo")
            undo_button.on_click(self.re_undo_button_clicked)
            redo_button = widgets.Button(description="Redo")
            redo_button.on_click(self.re_undo_button_clicked)
            init_button = widgets.Button(description="Set initial")
            init_button.on_click(lambda _: self.reset_to_init_params())

            button_list.append(undo_button)
            button_list.append(redo_button)
            button_list.append(init_button)

        combo_widget = [
            widgets.HBox([v, w])
            for v, w in zip(list(self.param_widgets.values()), checkboxes)
        ]
        if is_in_notebook():
            display(
                widgets.VBox(
                    [
                        widgets.HBox(
                            combo_widget + button_list,
                            layout=widgets.Layout(
                                width="100%",
                                display="inline-flex",
                                flex_flow="row wrap",
                            ),
                        ),
                        self.fig,
                    ]
                )
            )

    def fit_function(
        self,
        params: Parameters,
        lbda: npt.NDArray,
        rhor: npt.NDArray,
        rhoi: npt.NDArray,
    ) -> npt.NDArray:
        """The fit function to minimize the fitting problem

        Args:
            params (Parameters):
                The lmfit fitting Parameters to construct the simulation
            lbda (npt.NDArray): Wavelengths in nm
            rhor (npt.NDArray): The real part of the experimental rho
            rhoi (npt.NDArray): The imaginary part of the experimental rho

        Returns:
            npt.NDArray:
                Residual between the calculation with
                current parameters and experimental data
        """
        result = self.model(lbda, params)

        resid_rhor = rhor - result.rho.real
        resid_rhoi = rhoi - result.rho.imag

        return np.concatenate((resid_rhor, resid_rhoi))

    def fit(self, method="leastsq"):
        """Execute lmfit with the current fitting parameters

        Args:
            method (str, optional): The fitting method to use.
                                    Any method supported by scipys curve_fit is allowed.
                                    Defaults to 'leastsq'.

        Returns:
            Result: The fitting result
        """
        rho = calc_rho(self.exp_data)
        res = minimize(
            self.fit_function,
            self.params,
            args=(rho.index.to_numpy(), rho.values.real, rho.values.imag),
            method=method,
        )

        self.fitted_params = res.params
        return res

    def plot(self) -> go.Figure:
        """Plot the fit results as Psi/Delta"""
        fit_result = self.model(self.exp_data.index.to_numpy(), self.fitted_params)

        return go.FigureWidget(
            pd.concat(
                [
                    self.exp_data,
                    pd.DataFrame(
                        {"Ψ_fit": fit_result.psi, "Δ_fit": fit_result.delta},
                        index=self.exp_data.index,
                    ),
                ]
            ).plot(backend="plotly")
        )

    def plot_rho(self) -> go.Figure:
        """Plot the fit results as Rho"""
        rho = calc_rho(self.exp_data)
        fit_result = self.model(rho.index.to_numpy(), self.fitted_params)

        return go.FigureWidget(
            pd.DataFrame(
                {
                    "ρr": rho.apply(lambda x: x.real),
                    "ρi": rho.apply(lambda x: x.imag),
                    "ρcr": fit_result.rho.real,
                    "ρci": fit_result.rho.imag,
                },
                index=rho.index,
            ).plot(backend="plotly")
        )

    def get_model_data(
        self, params: Parameters = None, repr: str = "psi-delta", append_exp_data=False
    ) -> pd.DataFrame:
        """Gets the data from the provided modle with the provided parameters.
        If no parameters are provided, the fitted parameters are used
        (which default to the initial parameters if no fit has been triggered).

        Args:
            params (Parameters, optional):
                The parameters to calculate the model with.
                If not provided, the fitted parameters are used.
                Defaults to None.
            repr (str, optional):
                The representation of the model and experiment data.
                Valid values are 'psi-delta' or 'rho'. Defaults to 'psi-delta'.
            append_exp_data (bool, optional):
                Appends the experimental data if set to True.
                Defaults to False.

        Raises:
            ValueError: Raised if the representation is not a valid value.

        Returns:
            pd.DataFrame: The model results
        """
        if repr not in ["psi-delta", "rho"]:
            raise ValueError(f"Unexpected representation: {repr}")

        if params is None:
            fit_result = self.model(self.exp_data.index.to_numpy(), self.fitted_params)
            desc = "fit"
        else:
            fit_result = self.model(self.exp_data.index.to_numpy(), params)
            desc = "model"

        exp_data = {"psi-delta": self.exp_data, "rho": calc_rho(self.exp_data)}

        sim_data = {
            "psi-delta": pd.DataFrame(
                {f"Ψ_{desc}": fit_result.psi, f"Δ_{desc}": fit_result.delta},
                index=self.exp_data.index,
            ),
            "rho": pd.DataFrame(
                {f"ρ_{desc}": fit_result.rho}, index=self.exp_data.index
            ),
        }

        if append_exp_data:
            return pd.concat([exp_data[repr], sim_data[repr]], axis=1)

        return sim_data[repr]

    def __init__(
        self,
        exp_data: pd.DataFrame,
        params: Parameters,
        model: Callable[[npt.NDArray, Parameters], Result],
        angle: float = 70,
        **kwargs,
    ) -> None:
        """Intialize the psi/delta fitting class

        Args:
            exp_data (pd.DataFrame):
                The dataframe containing an experimental mueller matrix.
                It should contain 2 columns with labels Ψ and Δ.
            params (Parameters): Fitting start parameters
            model (Callable[[npt.NDArray, Parameters], Result]):
                A function taking wavelengths as first
                parameter and fitting parameters as second,
                which returns a pyEllis Result object.
                This function contains the actual model which should be fitted.
            angle (float, optional): The angle of incident of the measurement.
                                     Used to calculate the Pseudo-Dielectric function.
                                     Defaults to 70.
        """
        super().__init__()
        self.model = model
        self.exp_data = exp_data
        self.params = params
        self.fitted_params = params.copy()
        self.angle = angle
        self.param_widgets = {}
        self.selector = widgets.Dropdown()
        self.last_params = None
        self.initial_params = params.copy()
        self.fit_kwargs = kwargs
        self.fig = go.FigureWidget(
            pd.concat(
                [
                    exp_data,
                    pd.DataFrame(
                        {
                            "Ψ_calc": model(exp_data.index, params).psi,
                            "Δ_calc": model(exp_data.index, params).delta,
                        },
                        index=exp_data.index,
                    ),
                ]
            ).plot(backend="plotly")
        )
        self.fig.update_layout(yaxis_title="Ψ/Δ (°)", xaxis_title="Wavelength (nm)")

        self.update_dict = {
            "Psi/Delta": self.set_psi_delta,
            "Rho": self.set_rho,
            "Pseudo Diel.": self.set_pseudo_diel,
            "Residual": self.set_residual,
        }

        self.create_widgets()


def fit(
    exp_data: pd.DataFrame, params: Parameters, angle: float = 70, **kwargs
) -> Callable[[npt.NDArray, Parameters], Result]:
    """A parameters decorator for fitting psi/delta valus. Displays an ipywidget float box for
    each fitting parameter and an interactive plot to estimate parameters.

    Args:
        exp_data (pd.DataFrame): The dataframe containing an experimental mueller matrix.
                                 It should contain 2 columns with labels Ψ and Δ.
        params (Parameters): Fitting start parameters
        angle (float, optional): The angle of incident of the measurement.
                                 Used to calculate the Pseudo-Dielectric function.
                                 Defaults to 70.

    Returns:
        Callable[[npt.NDArray, Parameters], Result]:
            fitting parameters as second,
            which returns a pyEllis Result object.
            This function contains the actual model which
            should be fitted and is automatically
            provided when used as a decorator.
    """
    return lambda model: FitRho(exp_data, params, model, angle, **kwargs)



# Encoding: utf-8
"""Cauchy dispersion."""

import numpy.typing as npt

from .base_dispersion import IndexDispersion


class Cauchy(IndexDispersion):
    r"""Cauchy dispersion.

    Single parameters:
        :n0: Defaults to 1.5.
        :n1: Defaults to 0. Unit in nm\ :sup:`2`.
        :n2: Defaults to 0. Unit in nm\ :sup:`4`.
        :k0: Defaults to 0.
        :k1: Defaults to 0. Unit in nm\ :sup:`2`.
        :k2: Defaults to 0. Unit in nm\ :sup:`4`.

    Repeated parameters:
        --

    Output:
        .. math::
            \varepsilon^{1/2}(\lambda) =
            \boldsymbol{n_0} + 100  \boldsymbol{n_1}/\lambda^2 +
            10^7 \boldsymbol{n_2}/\lambda^4
            + i (\boldsymbol{k_0} + 100 \boldsymbol{k_1}/\lambda^2
            + 10^7 \boldsymbol{k_2}/\lambda^4)
    """

    single_params_template = {"n0": 1.5, "n1": 0, "n2": 0, "k0": 0, "k1": 0, "k2": 0}
    rep_params_template = {}

    def refractive_index(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return (
            self.single_params.get("n0")
            + 1e2 * self.single_params.get("n1") / lbda**2
            + 1e7 * self.single_params.get("n2") / lbda**4
            + 1j
            * (
                self.single_params.get("k0")
                + 1e2 * self.single_params.get("k1") / lbda**2
                + 1e7 * self.single_params.get("k2") / lbda**4
            )
        )



# Encoding: utf-8
"""Pseudo dielectric function dispersion."""

from typing import Union
import numpy as np
import numpy.typing as npt
from scipy.interpolate import interp1d

from .base_dispersion import Dispersion, InvalidParameters


class PseudoDielectricFunction(Dispersion):
    r"""A pseudo dielectric function generated from experimental psi/delta values.
    Please note that the pseudo dielectric function can lead to unphysical behaviour, such
    as negative refractive indices or other spurious artifacts.
    Additionally, this formula is only valid for a bulk absorbing material and yields wrong
    results for layered materials.
    Therefore, it is preferable to use the pseudo dielectric function only as a helper for
    constructing other dispersion models.

    Single parameters:
        :angle: The measurement angle in degree under which the psi/delta values where obtained.
        :lbda: The wavelength region of the measurement data. Units in nm.
        :psi: The psi values of the measurement. Units in degree.
        :delta: The delta values of the measurement. Units in degree.

    Repeated parameters:
        --

    Output:
        .. math::
            \varepsilon(\lambda) = \sin^2 \left( \Theta \right) \cdot
            \left( 1 + \tan^2 (\Theta) \left( \frac{1 - \rho}{1 + \rho} \right) \right)

        With

        .. math::
            \rho = \tan(\Psi) \cdot \exp (-i \Delta)

        :math:`\Theta` is the angle of incidence.
    """

    single_params_template = {"angle": None, "lbda": None, "psi": None, "delta": None}
    rep_params_template = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        rho = np.tan(np.deg2rad(self.single_params.get("psi"))) * np.exp(
            -1j * np.deg2rad(self.single_params.get("delta"))
        )
        theta = self.single_params.get("angle") * np.pi / 180
        eps = np.sin(theta) ** 2 * (
            1 + np.tan(theta) ** 2 * ((1 - rho) / (1 + rho)) ** 2
        )

        self.interpolation = interp1d(
            self.single_params.get("lbda"),
            eps,
            kind="cubic",
        )

    def __add__(self, _: Union[int, float, "Dispersion"]) -> "DispersionSum":
        raise NotImplementedError("Adding of tabular dispersions is not yet supported")

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        return self.interpolation(lbda)



"""Tests for dispersion models"""

import os
from shutil import copytree, rmtree

import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal
from pytest import fixture, raises

import elli
from elli.dispersions.base_dispersion import InvalidParameters


@fixture
def datadir(tmpdir, request):
    """Fixture for providing the dispersion prototype folder"""
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)

    if os.path.isdir(test_dir):
        rmtree(tmpdir)
        copytree(test_dir, str(tmpdir))

    return tmpdir


def test_constant_refr_index():
    """Tests constant refractive index"""
    disp = elli.ConstantRefractiveIndex(n=3)
    diel = disp.get_dielectric_df()
    eps0 = diel.loc[:, "ϵ1"].values
    eps1 = diel.loc[:, "ϵ2"].values

    assert_array_equal(np.ones(801) * 3**2, eps0)
    assert_array_equal(np.zeros(801), eps1)


def test_epsilon_inf():
    """Tests epsilon infinity / constant dielectric function"""
    disp = elli.EpsilonInf(eps=3)
    diel = disp.get_dielectric_df()
    eps0 = diel.loc[:, "ϵ1"].values
    eps1 = diel.loc[:, "ϵ2"].values

    assert_array_equal(np.ones(801) * 3, eps0)
    assert_array_equal(np.zeros(801), eps1)


def test_regression_dispersions_default(datadir):
    """Test dispersions against their prior values"""
    dispersions = [
        "Cauchy",
        "CauchyUrbach",
        "DrudeEnergy",
        "DrudeResistivity",
        "Gaussian",
        "LorentzEnergy",
        "LorentzLambda",
        "Poles",
        "Sellmeier",
        "Tanguy",
        "TaucLorentz",
    ]
    lbda = np.linspace(400, 1000, 500)

    for dispersion in dispersions:
        prototype = pd.read_csv(datadir.join(f"{dispersion}_default.csv"), index_col=0)
        disp = elli.DispersionFactory.get_dispersion(dispersion)

        assert_frame_equal(prototype, disp.get_dielectric_df(lbda))


def test_regression_dispersions_custom_values(datadir):
    """Test dispersions against their prior values"""
    dispersions = [
        {
            "name": "Cauchy",
            "single_params": {
                "n0": 1.5,
                "n1": 0.3,
                "n2": 0.05,
                "k0": 0.6,
                "k1": 0.2,
                "k2": 0.1,
            },
            "rep_params": [],
        },
        {
            "name": "CauchyUrbach",
            "single_params": {
                "n0": 1.5,
                "B": 0.005,
                "C": 0.0001,
                "D": 0.0001,
                "Eg": 3,
                "Eu": 0.4,
            },
            "rep_params": [],
        },
        {
            "name": "DrudeEnergy",
            "single_params": {"A": 100, "gamma": 0.5},
            "rep_params": [],
        },
        {
            "name": "DrudeResistivity",
            "single_params": {"rho_opt": 100, "tau": 1e-2},
            "rep_params": [],
        },
        {
            "name": "LorentzLambda",
            "single_params": {},
            "rep_params": [
                {"A": 100, "lambda_r": 500, "gamma": 10},
                {"A": 150, "lambda_r": 300, "gamma": 20},
                {"A": 300, "lambda_r": 750, "gamma": 50},
            ],
        },
        {
            "name": "LorentzEnergy",
            "single_params": {},
            "rep_params": [
                {"A": 100, "E": 3, "gamma": 0.1},
                {"A": 150, "E": 1.5, "gamma": 0.05},
                {"A": 300, "E": 0.3, "gamma": 0.02},
            ],
        },
        {
            "name": "Gaussian",
            "single_params": {},
            "rep_params": [
                {"A": 100, "E": 3, "sigma": 0.1},
                {"A": 150, "E": 1.5, "sigma": 0.05},
                {"A": 300, "E": 0.3, "sigma": 0.02},
            ],
        },
        {
            "name": "TaucLorentz",
            "single_params": {"Eg": 2},
            "rep_params": [
                {"A": 100, "E": 2.5, "C": 0.1},
                {"A": 150, "E": 3, "C": 0.05},
                {"A": 300, "E": 4.5, "C": 0.02},
            ],
        },
        {
            "name": "Tanguy",
            "single_params": {
                "A": 1,
                "d": 2,
                "gamma": 0.1,
                "R": 0.1,
                "Eg": 2,
                "a": 1,
                "b": 0,
            },
            "rep_params": [],
        },
        {
            "name": "Poles",
            "single_params": {"A_ir": 100, "A_uv": 100, "E_uv": 4},
            "rep_params": [],
        },
        {
            "name": "Table",
            "single_params": {
                "lbda": np.linspace(400, 1000, 100),
                "n": np.linspace(1, 1.5, 100) + 1j * np.linspace(0, 1, 100),
            },
            "rep_params": [],
        },
        {
            "name": "TableEpsilon",
            "single_params": {
                "lbda": np.linspace(400, 1000, 100),
                "epsilon": np.linspace(1, 1.5, 100) + 1j * np.linspace(0, 1, 100),
            },
            "rep_params": [],
        },
        {
            "name": "PseudoDielectricFunction",
            "single_params": {
                "angle": 70,
                "lbda": np.linspace(400, 1000, 100),
                "psi": np.linspace(0, 90, 100),
                "delta": np.linspace(0, 90, 100),
            },
            "rep_params": [],
        },
    ]
    lbda = np.linspace(400, 1000, 500)

    for dispersion in dispersions:
        disp_name = dispersion.get("name")
        prototype = pd.read_csv(
            datadir.join(f"{disp_name}_custom_values.csv"), index_col=0
        )
        disp = elli.DispersionFactory.get_dispersion(
            disp_name, **dispersion.get("single_params")
        )
        for rep_param in dispersion.get("rep_params"):
            disp.add(**rep_param)

        assert_frame_equal(prototype, disp.get_dielectric_df(lbda))


def test_fail_parameterless_pseudo_dielectric():
    """Checks whether calling the pseudo dielectric function fails when called
    without single parameters"""
    with raises(InvalidParameters):
        _ = elli.DispersionFactory.get_dispersion("PseudoDielectricFunction")


def test_correct_reconstruction_with_pseudo_dielectric():
    """Checks whether the pseudo dielectric function
    reconstructs a predefined model"""

    gaussian = elli.Gaussian().add(A=10, E=2, sigma=0.3)
    lbda = gaussian.get_dielectric_df().index.values

    result = elli.Structure(
        elli.AIR,
        [],
        gaussian.get_mat(),
    ).evaluate(lbda, 70, solver=elli.Solver2x2)

    check_lbda = np.linspace(200, 1000, 5000)
    assert_frame_equal(
        elli.PseudoDielectricFunction(
            angle=70, lbda=lbda, psi=result.psi, delta=result.delta
        ).get_dielectric_df(check_lbda),
        gaussian.get_dielectric_df(check_lbda),
    )



"""This files generates files for the current values of the dispersions"""

import elli
import numpy as np


def execute_and_save(dispersion, identifier, rep_params, *args, **kwargs):
    lbda = np.linspace(400, 1000, 500)
    fname = f"test_dispersions/{dispersion}_{identifier}.csv"

    disp = elli.DispersionFactory.get_dispersion(dispersion, *args, **kwargs)
    for rep_param in rep_params:
        disp.add(**rep_param)
    disp.get_dielectric_df(lbda).to_csv(fname)


def loop_dispersions_default():
    dispersions = [
        "Cauchy",
        "CauchyUrbach",
        "DrudeEnergy",
        "DrudeResistivity",
        "Gaussian",
        "LorentzEnergy",
        "LorentzLambda",
        "Poles",
        "Sellmeier",
        "Tanguy",
        "TaucLorentz",
        "Table",
        "TableEpsilon",
    ]

    for dispersion in dispersions:
        execute_and_save(dispersion, "default", [])


def loop_dispersions_custom_values():
    dispersions = [
        {
            "name": "Cauchy",
            "single_params": {
                "n0": 1.5,
                "n1": 0.3,
                "n2": 0.05,
                "k0": 0.6,
                "k1": 0.2,
                "k2": 0.1,
            },
            "rep_params": [],
        },
        {
            "name": "CauchyUrbach",
            "single_params": {
                "n0": 1.5,
                "B": 0.005,
                "C": 0.0001,
                "D": 0.0001,
                "Eg": 3,
                "Eu": 0.4,
            },
            "rep_params": [],
        },
        {
            "name": "DrudeEnergy",
            "single_params": {"A": 100, "gamma": 0.5},
            "rep_params": [],
        },
        {
            "name": "DrudeResistivity",
            "single_params": {"rho_opt": 100, "tau": 1e-2},
            "rep_params": [],
        },
        {
            "name": "LorentzLambda",
            "single_params": {},
            "rep_params": [
                {"A": 100, "lambda_r": 500, "gamma": 10},
                {"A": 150, "lambda_r": 300, "gamma": 20},
                {"A": 300, "lambda_r": 750, "gamma": 50},
            ],
        },
        {
            "name": "LorentzEnergy",
            "single_params": {},
            "rep_params": [
                {"A": 100, "E": 3, "gamma": 0.1},
                {"A": 150, "E": 1.5, "gamma": 0.05},
                {"A": 300, "E": 0.3, "gamma": 0.02},
            ],
        },
        {
            "name": "Gaussian",
            "single_params": {},
            "rep_params": [
                {"A": 100, "E": 3, "sigma": 0.1},
                {"A": 150, "E": 1.5, "sigma": 0.05},
                {"A": 300, "E": 0.3, "sigma": 0.02},
            ],
        },
        {
            "name": "TaucLorentz",
            "single_params": {"Eg": 2},
            "rep_params": [
                {"A": 100, "E": 2.5, "C": 0.1},
                {"A": 150, "E": 3, "C": 0.05},
                {"A": 300, "E": 4.5, "C": 0.02},
            ],
        },
        {
            "name": "Tanguy",
            "single_params": {
                "A": 1,
                "d": 2,
                "gamma": 0.1,
                "R": 0.1,
                "Eg": 2,
                "a": 1,
                "b": 0,
            },
            "rep_params": [],
        },
        {
            "name": "Poles",
            "single_params": {"A_ir": 100, "A_uv": 100, "E_uv": 4},
            "rep_params": [],
        },
        {
            "name": "Table",
            "single_params": {
                "lbda": np.linspace(400, 1000, 100),
                "n": np.linspace(1, 1.5, 100) + 1j * np.linspace(0, 1, 100),
            },
            "rep_params": [],
        },
        {
            "name": "TableEpsilon",
            "single_params": {
                "lbda": np.linspace(400, 1000, 100),
                "epsilon": np.linspace(1, 1.5, 100) + 1j * np.linspace(0, 1, 100),
            },
            "rep_params": [],
        },
        {
            "name": "PseudoDielectricFunction",
            "single_params": {
                "angle": 70,
                "lbda": np.linspace(400, 1000, 100),
                "psi": np.linspace(0, 90, 100),
                "delta": np.linspace(0, 90, 100),
            },
            "rep_params": [],
        },
    ]

    for dispersion in dispersions:
        execute_and_save(
            dispersion.get("name"),
            "custom_values",
            dispersion.get("rep_params"),
            **dispersion.get("single_params"),
        )


if __name__ == "__main__":
    loop_dispersions_default()
    loop_dispersions_custom_values()



# Encoding: utf-8
"""Tauc-Lorentz dispersion law. Model by Jellison and Modine."""

import numpy as np
import numpy.typing as npt
from numpy.lib.scimath import sqrt

from ..utils import conversion_wavelength_energy
from .base_dispersion import Dispersion


class TaucLorentz(Dispersion):
    """Tauc-Lorentz dispersion law. Model by Jellison and Modine.

    Single parameters:
        :Eg: Bandgap energy (eV). Defaults to 1.

    Repeated parameters:
        :A: Strength of the absorption. Typically 10 < A < 200. Defaults to 20.
        :E: Lorentz resonance energy (eV). Always keep E > Eg!!. Defaults to 1.5.
        :C: Lorentz broadening (eV). Typically 0 < Ci < 10. Defaults to 1.

    Output:
        The Tauc lorentz dispersion. Please refer to the references for a full formula.

    References:
        * G.E. Jellision and F.A. Modine, Appl. Phys. Lett. 69 (3), 371-374 (1996)
        * Erratum, G.E. Jellison and F.A. Modine, Appl. Phys. Lett 69 (14), 2137 (1996)
        * H. Chen, W.Z. Shen, Eur. Phys. J. B. 43, 503-507 (2005)
    """

    single_params_template = {"Eg": 1}
    rep_params_template = {"A": 20, "E": 1.5, "C": 1}

    @staticmethod
    def eps1(E, Eg, Ai, Ei, Ci):
        gamma2 = sqrt(Ei**2 - Ci**2 / 2) ** 2
        alpha = sqrt(4 * Ei**2 - Ci**2)
        aL = (Eg**2 - Ei**2) * E**2 + Eg**2 * Ci**2 - Ei**2 * (Ei**2 + 3 * Eg**2)
        aA = (E**2 - Ei**2) * (Ei**2 + Eg**2) + Eg**2 * Ci**2
        zeta4 = (E**2 - gamma2) ** 2 + alpha**2 * Ci**2 / 4

        # fmt: off
        return (
            Ai*Ci*aL/2.0/np.pi/zeta4/alpha/Ei*np.log((Ei**2 + Eg**2 + alpha*Eg)/(Ei**2 + Eg**2 - alpha*Eg)) - \
            Ai*aA/np.pi/zeta4/Ei*(np.pi - np.arctan((2.0*Eg + alpha)/Ci) + np.arctan((alpha - 2.0*Eg)/Ci)) + \
            2.0*Ai*Ei*Eg/np.pi/zeta4/alpha*(E**2 - gamma2)*(np.pi + 2.0*np.arctan(2.0/alpha/Ci*(gamma2 - Eg**2))) - \
            Ai*Ei*Ci*(E**2 + Eg**2)/np.pi/zeta4/E*np.log(abs(E - Eg)/(E + Eg)) + \
            2.0*Ai*Ei*Ci*Eg/np.pi/zeta4 * \
            np.log(abs(E - Eg) * (E + Eg) / sqrt((Ei**2 - Eg**2)**2 + Eg**2 * Ci**2))
        )
        # fmt: on

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        energy = conversion_wavelength_energy(lbda)
        energy_g = self.single_params.get("Eg")
        return sum(
            (
                1j
                * (
                    c.get("A")
                    * c.get("E")
                    * c.get("C")
                    * (energy - energy_g) ** 2
                    / ((energy**2 - c.get("E") ** 2) ** 2 + c.get("C") ** 2 * energy**2)
                    / energy
                )
                * np.heaviside(energy - energy_g, 0)
                + self.eps1(energy, energy_g, c.get("A"), c.get("E"), c.get("C"))
            )
            for c in self.rep_params
        )



"""
TiO2/SiO2 Bragg mirror
======================
"""

# %%
# Example of a TiO2/SiO2 Bragg mirror with 8.5 periods
# ----------------------------------------------------
#
# Authors: O. Castany, M.Müller

# %%
import elli
import elli.plot as elliplot
import matplotlib.pyplot as plt
import numpy as np

np.set_printoptions(suppress=True, precision=3)

# %%
# Material definition
# -------------------
# We define air as incidence material and glass as exit material.
# SiO2 and TiO2 are defined by simplified dispersion relations.
air = elli.AIR
glass = elli.ConstantRefractiveIndex(1.5).get_mat()

n_SiO2 = 1.47
n_TiO2 = 2.23 + 1j * 5.2e-4

SiO2 = elli.ConstantRefractiveIndex(n_SiO2).get_mat()
TiO2 = elli.ConstantRefractiveIndex(n_TiO2).get_mat()

# %%
# Create layers and structure
# ---------------------------
# The SiO2 and TiO2 layers are set to the thickness of an
# quarterwaveplate of the respective material at 1550 nm.
#
# The layers are then stacked alternatingly and put into the
# complete structure with air and the glass substrate.
lbda0 = 1550

d_SiO2 = elli.get_qwp_thickness(SiO2, lbda0)
d_TiO2 = elli.get_qwp_thickness(TiO2, lbda0)

print("Thickness of the SiO2 QWP: {} nm".format(d_SiO2))
print("Thickness of the TiO2 QWP: {} nm".format(d_TiO2))

L_SiO2 = elli.Layer(SiO2, d_SiO2)
L_TiO2 = elli.Layer(TiO2, d_TiO2)

# Repeated layers: 8.5 periods
layerstack = elli.RepeatedLayers([L_TiO2, L_SiO2], 8, 0, 1)

s = elli.Structure(air, [layerstack], glass)

# %%
# Calculation
# -----------
(lbda1, lbda2) = (1100, 2500)
lbda_list = np.linspace(lbda1, lbda2, 200)

data = s.evaluate(lbda_list, 0)

R = data.R
T = data.T


# %%
# Structure Graph
# ---------------
# Schema of the variation of the refractive index in z-direction.
elliplot.draw_structure(s)

# %%
# Reflection and Transmission Graph
# ---------------------------------
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
ax.plot(lbda_list, T, label="$T$")
ax.plot(lbda_list, R, label="$R$")
ax.legend(loc="center right")
ax.set_xlabel(r"Wavelength $\lambda$ (nm)")
ax.set_ylabel(r"Power reflection $R$ or transmission $T$")
ax.set_title(r"Bragg mirror: Air/{TiO$_2$/SiO$_2$}x8/TiO$_2$/Glass")
plt.show()



# Encoding: utf-8
"""The virtual representation of a sample is finally build
using the Structure class.

A :class:`Structure` consists of one IsotropicMaterial used as semi-infinite
entry material and a second arbitrary Material as exit material.
A list of Layers can be added as desired in between these materials,
to create a 1D model of layered media.

The basic :class:`Layer` consists of a material and an assigned thickness.
An arbitrary sequence of layers can be stacked and repeated by :class:`RepeatedLayers`,
to create Bragg-mirror or multiple quantum well structures.

There are also classes to approximate layers with varying properties along the z-axis
(inhomogeneous layers) by creating multiple thinner homogeneous slices:

* :class:`TwistedLayer` is able to represent rotating materials, like twisted nematic materials.
* :class:`VaryingMixtureLayer` takes an :class:`MixtureMaterial<elli.materials.MixtureMaterial>` and uses a gradient as mixture fraction.
"""

from abc import ABC, abstractmethod
from typing import Callable, List, Tuple

import numpy as np
import numpy.typing as npt

from .experiment import Experiment
from .materials import IsotropicMaterial, Material, MixtureMaterial
from .result import Result
from .solver import Solver
from .solver4x4 import Solver4x4
from .utils import E_Z, rotation_v_theta


class AbstractLayer(ABC):
    """Abstract class for a layer."""

    @abstractmethod
    def get_permittivity_profile(
        self, lbda: npt.ArrayLike
    ) -> List[Tuple[float, npt.NDArray]]:
        """Returns the permittivity profile of the layer for the given wavelengths.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            List[Tuple[float, npt.NDArray]]:
                Returns list of tuples [(thickness, dielectric tensor), ...]
        """


class RepeatedLayers(AbstractLayer):
    """Repeated structure of layers."""

    repetitions = None  # Number of repetitions
    before = None  # additional layers before the first period
    after = None  # additional layers after the last period
    layers = None  # layers to repeat

    def __init__(
        self,
        layers: List[AbstractLayer],
        repetitions: int,
        before: int = 0,
        after: int = 0,
    ) -> None:
        """Create a repeated structure of layers.

        Example : For layers [1,2,3] with n=2, before=1 and after=0, the structure will be 3123123.

        Args:
            layers (List[AbstractLayer]): List of the repeated layers, starting from z=0
            repetitions (int): Number of repetitions
            before (int, optional):
                Number of additional layers before the first period. Defaults to 0.
            after (int, optional): Number of additional layers after the last period. Defaults to 0.
        """
        self.set_repetitions(repetitions, before, after)
        self.set_layers(layers)

    def set_repetitions(
        self, repetitions: int, before: int = 0, after: int = 0
    ) -> None:
        """Defines the number of repetitions and the first and last layers.

        Example : For layers [1,2,3] with n=2, before=1 and after=0, the structure will be 3123123.

        Args:
            repetitions (int): Number of repetitions
            before (int, optional):
                Number of additional layers before the first period. Defaults to 0.
            after (int, optional): Number of additional layers after the last period. Defaults to 0.
        """
        if repetitions < 1:
            raise ValueError("Number of repetitions need to be at least 1.")

        if before < 0:
            raise ValueError("Before value needs to be 0 or more.")

        if after < 0:
            raise ValueError("Divisions needs to be 0 or more.")

        self.repetitions = repetitions
        self.before = before
        self.after = after

    def set_layers(self, layers: List[AbstractLayer]) -> None:
        """Set list of layers.

        Args:
            layers (List[AbstractLayer]): List of the repeated layers, starting from z=0
        """
        if not isinstance(layers, list):
            raise TypeError("Provide a list of layers.")
        for layer in layers:
            if not isinstance(layer, AbstractLayer):
                raise TypeError(
                    f"Expected a list of Layer objects but found an item of type {type(layer)}."
                )

        self.layers = layers

    def get_permittivity_profile(
        self, lbda: npt.ArrayLike
    ) -> List[Tuple[float, npt.NDArray]]:
        """Returns the permittivity profile of the layer for the given wavelengths.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            List[Tuple[float, npt.NDArray]]:
                Returns list of tuples [(thickness, dielectric tensor), ...]
        """
        layers = []
        for layer in self.layers:
            layers += layer.get_permittivity_profile(lbda)

        if self.before > 0:
            before = layers[-self.before :]
        else:
            before = []
        return before + self.repetitions * layers + layers[: self.after]


class Layer(AbstractLayer):
    """Homogeneous layer of dielectric material."""

    thickness = None

    def __init__(self, material: Material, thickness: float) -> None:
        """New layer of material 'material', with thickness 'thickness'

        Args:
            material (Material): Material object
            thickness (float): Thickness of layer (in nm)
        """
        self.set_material(material)
        self.set_thickness(thickness)

    def set_thickness(self, thickness: float) -> None:
        """Defines the thickness of the layer in nm.

        Args:
            thickness (float): Thickness of the layer in nm.
        """
        if thickness < 0:
            raise ValueError("Thickness value can't be negative.")

        self.thickness = thickness

    def set_material(self, material: Material) -> None:
        """Defines the material for the layer.

        Args:
            material (Material): Material object
        """
        if not isinstance(material, Material):
            raise TypeError(
                f"Expected material to be an Material object but found type {type(material)}."
            )

        self.material = material

    def get_permittivity_profile(
        self, lbda: npt.ArrayLike
    ) -> List[Tuple[float, npt.NDArray]]:
        """Returns the permittivity profile of the layer for the given wavelengths.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            List[Tuple[float, npt.NDArray]]:
                Returns a list containing one tuple [(thickness, dielectric tensor)]
        """
        return [(self.thickness, self.material.get_tensor(lbda))]


#########################################################
# Inhomogeneous Layers


class InhomogeneousLayer(AbstractLayer):
    """Abstract base class for inhomogeneous layers with varying properties in z-direction."""

    thickness = None
    material = None
    div = None

    def set_thickness(self, thickness: float) -> None:
        """Defines the thickness of the layer in nm.

        Args:
            thickness (float): Thickness of the layer in nm.
        """
        if thickness < 0:
            raise ValueError("Thickness value can't be negative.")

        self.thickness = thickness

    def set_material(self, material: Material) -> None:
        """Defines the material for the layer.

        Args:
            material (Material): Material object
        """
        if not isinstance(material, Material):
            raise TypeError(
                f"Expected material to be an Material object but found type {type(material)}."
            )

        self.material = material

    def set_divisions(self, div: int) -> None:
        """Defines the number of slices to simulate the layer.

        Args:
            div (int): Number of slices for the layer
        """
        if div < 1:
            raise ValueError("Number of slices need to be at least 1.")

        self.div = div

    def get_slices(self) -> npt.NDArray:
        """Returns z slicing with the position relative to this layer, not to the whole structure.

        Returns:
            npt.NDArray: array of 'z' positions [z0, z1,... , zmax], with z0 = 0 and zmax = z{d+1}
        """
        return np.linspace(0, self.thickness, self.div + 1)

    @abstractmethod
    def get_tensor(self, z: float, lbda: npt.ArrayLike) -> npt.NDArray:
        """Returns permittivity tensor matrix for position 'z'."""

    def get_permittivity_profile(self, lbda: npt.ArrayLike) -> List:
        """Returns the permittivity profile of the layer for the given wavelengths.
        The tensor is evaluated in the middle of each slice.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            List[Tuple[float, npt.NDArray]]:
                Returns list of tuples [(d1, epsilon1), (d2, epsilon2), ...]
        """
        z = self.get_slices()
        h = np.diff(z)
        zmid = (z[:-1] + z[1:]) / 2.0
        tensor = [self.get_tensor(z, lbda) for z in zmid]
        return list(zip(h, tensor))


class TwistedLayer(InhomogeneousLayer):
    """Twisted layer.
    The material gets rotated around the z axis."""

    def __init__(
        self, material: Material, thickness: float, div: int, angle: float
    ) -> None:
        """Creates a layer with a twisted material.

        Args:
            material (Material): Material object
            thickness (float): Thickness of layer (in nm)
            div (int): Number of slices for the layer
            angle (float): rotation angle over the distance 'd' (in degrees)
        """
        self.set_material(material)
        self.set_thickness(thickness)
        self.set_divisions(div)
        self.set_angle(angle)

    def set_angle(self, angle: float) -> None:
        """Defines the total twist angle of this layer.

        Args:
            angle (float): Rotation angle over the thickness 'd' of the layer (in degrees)
        """
        self.angle = angle

    def get_tensor(self, z: float, lbda: npt.ArrayLike) -> npt.NDArray:
        """Gets permittivity tensor matrix for position 'z' and wavelength 'lbda'.

        Args:
            z (float): Position in the layer (in nm)
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            npt.NDArray: Permittivity tensor for position 'z' and wavelength 'lbda'.
        """
        epsilon = self.material.get_tensor(lbda)
        m_r = rotation_v_theta(E_Z, self.angle * z / self.thickness)
        return m_r @ epsilon @ m_r.T


class VaryingMixtureLayer(InhomogeneousLayer):
    """Mixture layer, with varying fraction dependent on z Position.

    Note: The set fraction of the mixture material will be ignored
    and replaced by the result of the fraction modulation function.
    """

    def __init__(
        self,
        material: MixtureMaterial,
        thickness: float,
        div: int,
        fraction_modulation: Callable[[float], float] = lambda x: x,
    ) -> None:
        """
        Args:
            material (MixtureMaterial): MixtureMaterial object
            d (float): Thickness of layer (in nm)
            div (int): Number of slices for the layer
            fraction_modulation (Callable[[float], float]):
                Function to modify the fraction amount,
                takes float from 0 to 1 (top to bottom of layer),
                should return fraction at that level.
                Defaults to a linear profile
                (100% host material to 100% guest material).
        """
        self.set_material(material)
        self.set_thickness(thickness)
        self.set_divisions(div)
        self.set_fraction_modulation(fraction_modulation)

    def set_material(self, material: MixtureMaterial) -> None:
        """Defines the material for the varying mixture layer.

        Args:
            material (MixtureMaterial): Material object
        """
        if not isinstance(material, MixtureMaterial):
            raise TypeError(
                f"Expected material to be an MixtureMaterial object but found type {type(material)}."
            )

        self.material = material

    def set_fraction_modulation(
        self, fraction_modulation: Callable[[float], float] = lambda x: x
    ) -> None:
        """Sets function for variation of the mixture over the layer

        Args:
            fraction_modulation (Callable[[float], float]):
                Function to modify the fraction amount,
                takes float from 0 to 1 (top to bottom of layer),
                should return fraction at that level.
                Defaults to a linear profile
                (100% host material to 100% guest material).
        """
        self.fraction_modulation = fraction_modulation

    def get_tensor(self, z: float, lbda: npt.ArrayLike) -> npt.NDArray:
        """Gets permittivity tensor matrix for position 'z' and wavelength 'lbda'.

        Args:
            z (float): Position in the layer (in nm)
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            npt.NDArray: Permittivity tensor for position 'z' and wavelength 'lbda'.
        """
        epsilon = self.material.get_tensor_fraction(
            lbda, self.fraction_modulation(z / self.thickness)
        )
        return epsilon


#########################################################
# Structure Class


class Structure:
    """Description of the whole structure.

    Consists of:

    * front half-space (incident)
    * layer succession
    * back half-space (exit)
    """

    front_material = None
    back_material = None
    layers = []  # list of layers

    def __init__(
        self, front: IsotropicMaterial, layers: List[Layer], back: Material
    ) -> None:
        """
        Args:
            front (IsotropicMaterial): IsotropicMaterial used as front half space
            layers (List[Layer]): List of Layers, starting from z=0
            back (Material): Material used as back half space
        """
        self.set_front_material(front)
        self.set_layers(layers)
        self.set_back_material(back)

    def set_front_material(self, material: IsotropicMaterial) -> None:
        """Defines the front half-space material. Has to be isotropic.

        Args:
            material (IsotropicMaterial): IsotropicMaterial used as front half space
        """
        if not isinstance(material, Material):
            raise TypeError(
                f"Expected front material to be an Material object but found type {type(material)}."
            )

        self.front_material = material

    def set_back_material(self, material: Material) -> None:
        """Defines the back half-space material.

        Args:
            material (Material): Material used as back half space
        """
        if not isinstance(material, Material):
            raise TypeError(
                f"Expected back material to be an Material object but found type {type(material)}."
            )

        self.back_material = material

    def set_layers(self, layers: List[Layer]) -> None:
        """Sets sequence of layers.

        Args:
            layers (List[Layer]): List of Layers, starting from z=0
        """
        if not isinstance(layers, list):
            raise TypeError("Provide a list of layers.")
        for layer in layers:
            if not isinstance(layer, AbstractLayer):
                raise TypeError(
                    f"Expected a list of Layer objects but found an item of type {type(layer)}."
                )

        self.layers = layers

    def get_permittivity_profile(
        self, lbda: npt.ArrayLike
    ) -> List[Tuple[float, npt.NDArray]]:
        """Returns the permittivity profile of the complete structure for the given wavelengths.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).

        Returns:
            List[Tuple[float, npt.NDArray]]:
                Returns list of tuples [(thickness, dielectric tensor), ...]
        """
        permittivity_profile = [(np.inf, self.front_material.get_tensor(lbda))]

        for layer in self.layers:
            permittivity_profile.extend(layer.get_permittivity_profile(lbda))

        permittivity_profile.extend([(np.inf, self.back_material.get_tensor(lbda))])
        return permittivity_profile

    def evaluate(
        self,
        lbda: npt.ArrayLike,
        theta_i: float,
        solver: Solver = Solver4x4,
        **solver_kwargs,
    ) -> Result:
        """Return the Evaluation of the structure for the given parameters with standard settings.

        Args:
            lbda (npt.ArrayLike): Single value or array of wavelengths (in nm).
            theta_i (float): Incident angle of the experiment (in degrees).
            solver (Solver, optional): Choose which solver class is used. Defaults to Solver4x4.
            solver_kwargs (optional): Keyword arguments for the Solver can be appended as arguments.

        Returns:
            Result: Result of the experiment.
        """
        exp = Experiment(self, lbda, theta_i)
        return exp.evaluate(solver, **solver_kwargs)



# Encoding: utf-8
"""Constant refractive index."""

import numpy.typing as npt

from .base_dispersion import IndexDispersion


class ConstantRefractiveIndex(IndexDispersion):
    r"""Constant refractive index.

    Single parameters:
        :n: The constant value of the refractive index. Defaults to 1.

    Repeated parameters:
        --

    Output:
        .. math::
            \varepsilon(\lambda) = \boldsymbol{n}^2
    """

    single_params_template = {"n": 1}
    rep_params_template = {}

    def refractive_index(self, _: npt.ArrayLike) -> npt.NDArray:
        return self.single_params.get("n")



# Encoding: utf-8
"""Drude dispersion model with resistivity based parameters."""

import numpy as np
import numpy.typing as npt
import scipy.constants as sc

from ..utils import conversion_wavelength_energy
from .base_dispersion import Dispersion


class DrudeResistivity(Dispersion):
    r"""Drude dispersion model with resistivity based parameters.
    Drude models in the literature typically contain an additional epsilon infinity value.
    Use `EpsilonInf` to add this parameter or simply do DrudeEnergy() + eps_inf.

    Single parameters:
        :rho_opt: Optical resistivity. Defaults to 1. Unit in Ω-cm.
        :tau: Mean scattering time. Defaults to 1. Unit in s.

    Repeated parameters:
        --

    Output:

        .. math::
            \varepsilon(E) = \hbar / (\varepsilon_0  \cdot
            \boldsymbol{rho\_opt} \cdot \boldsymbol{tau} \cdot E^2
            - i \cdot \hbar \cdot E)

       where :math:`\hbar` is the planck constant divided by :math:`2\pi`
       and :math:`\varepsilon_0` is the vacuum dielectric permittivity.
    """

    single_params_template = {"rho_opt": 1, "tau": 1}
    rep_params_template = {}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        energy = conversion_wavelength_energy(lbda)
        hbar = sc.value("Planck constant in eV/Hz") / 2 / np.pi
        eps0 = sc.value("vacuum electric permittivity") * 1e-2

        return hbar**2 / (
            eps0
            * self.single_params.get("rho_opt")
            * (self.single_params.get("tau") * energy**2 - 1j * hbar * energy)
        )



"""Test loading optical models from nexus dispersive material files"""

import numpy as np
from numpy.testing import assert_array_almost_equal
import h5py

from fixtures import datadir  # pylint: disable=unused-import
from elli.dispersions.table_index import Table
from elli.importer.nexus import read_nexus_materials
from elli.dispersions.base_dispersion import IndexDispersionSum
from elli.dispersions.polynomial import Polynomial
from elli.dispersions.sellmeier import Sellmeier
from elli.dispersions.sellmeier_custom import SellmeierCustomExponent
from elli.materials import BiaxialMaterial, IsotropicMaterial, UniaxialMaterial


def test_read_nexus_without_errors(datadir):
    """All dispersive materials are loaded without errors"""
    for file in datadir.iterdir():
        read_nexus_materials(file)


def test_read_nexus_nk_table(datadir):
    """An isotropic nk table is loaded properly"""
    nk_file = datadir / "MoTe2-Beal.nxs"

    with h5py.File(nk_file, "r") as h5file:
        dispersion_path = "/entry/dispersion_x/dispersion_table_nk"
        lbda = h5file[f"{dispersion_path}/wavelength"][()]

        table_manual = Table(
            lbda=lbda,
            n=h5file[f"{dispersion_path}/refractive_index"][()],
        )

    nexus_material = read_nexus_materials(nk_file)["entry"]

    assert isinstance(nexus_material, IsotropicMaterial)

    assert_array_almost_equal(
        table_manual.get_dielectric(lbda),
        nexus_material.dispersion_x.get_dielectric(lbda),
    )


def test_read_nexus_uniaxial_material_with_table(datadir):
    """An uniaxial material with tabulated data is loaded correctly."""
    uniaxial_file = datadir / "MoSe2-Munkhbat-o.nxs"

    with h5py.File(uniaxial_file, "r") as h5file:
        dispersion_x_path = "/entry/dispersion_x/dispersion_table_nk"
        lbda = h5file[f"{dispersion_x_path}/wavelength"][()]

        table_x_manual = Table(
            lbda=lbda,
            n=h5file[f"{dispersion_x_path}/refractive_index"][()],
        )

        dispersion_z_path = "/entry/dispersion_z/dispersion_table_n"
        lbda = h5file[f"{dispersion_z_path}/wavelength"][()]

        table_z_manual = Table(
            lbda=lbda, n=h5file[f"{dispersion_z_path}/refractive_index"][()]
        )

    nexus_material = read_nexus_materials(uniaxial_file)["entry"]

    assert isinstance(nexus_material, UniaxialMaterial)

    assert_array_almost_equal(
        table_x_manual.get_dielectric(lbda),
        nexus_material.dispersion_x.get_dielectric(lbda),
    )

    assert_array_almost_equal(
        table_z_manual.get_dielectric(lbda),
        nexus_material.dispersion_z.get_dielectric(lbda),
    )


def test_read_nexus_for_sellmeier_table_sum(datadir):
    """A sum of sellmeier and a table is read correctly"""
    nexus = datadir / "Si-Chandler-Horowitz.nxs"

    def read_sellmeier(h5file: h5py.File):
        sellmeier_path = "/entry/dispersion_x/sellmeier"
        params_A = h5file[f"{sellmeier_path}/A/values"][()]
        params_B = h5file[f"{sellmeier_path}/B/values"][()]
        params_e1 = h5file[f"{sellmeier_path}/e1/values"][()]
        params_e2 = h5file[f"{sellmeier_path}/e2/values"][()]
        eps_inf = h5file["/entry/dispersion_x/sellmeier/eps_inf/value"][()]

        sellmeier = SellmeierCustomExponent()
        for A, B, e1, e2 in zip(params_A, params_B, params_e1, params_e2):
            sellmeier.add(A, e1, B, e2)

        return (sellmeier + eps_inf).as_index()

    with h5py.File(nexus, "r") as h5file:
        table_path = "/entry/dispersion_x/dispersion_table_k"
        lbda = h5file[f"{table_path}/wavelength"][()]

        table = Table(lbda=lbda, n=h5file[f"{table_path}/refractive_index"][()])

        dispersion = table + read_sellmeier(h5file)

    nexus_material = read_nexus_materials(nexus)["entry"]

    assert isinstance(nexus_material, IsotropicMaterial)
    assert isinstance(nexus_material.dispersion_x, IndexDispersionSum)

    assert_array_almost_equal(
        dispersion.get_dielectric(lbda),
        nexus_material.dispersion_x.get_dielectric(lbda),
    )


def test_read_nexus_sellmeier(datadir):
    """A sellmeier dispersion is read correctly"""
    nexus = datadir / "Si-Salzberg.nxs"
    lbda = np.linspace(1357, 11040, 100)

    with h5py.File(nexus, "r") as h5file:
        sellmeier_path = "/entry/dispersion_x/sellmeier"
        params_A = h5file[f"{sellmeier_path}/A/values"][()]
        params_B = h5file[f"{sellmeier_path}/B/values"][()]

        sellmeier = Sellmeier()
        for A, B in zip(params_A, params_B):
            sellmeier.add(A, B**2)

    nexus_material = read_nexus_materials(nexus)["entry"]

    assert isinstance(nexus_material, IsotropicMaterial)

    assert_array_almost_equal(
        sellmeier.get_dielectric(lbda), nexus_material.dispersion_x.get_dielectric(lbda)
    )


def test_read_nexus_biaxial_material(datadir):
    """A biaxial material is read correctly"""
    nexus = datadir / "WTe2-Munkhbat-alpha.nxs"

    with h5py.File(nexus, "r") as h5file:
        table_x_path = "/entry/dispersion_x/dispersion_table_nk"
        table_y_path = "/entry/dispersion_y/dispersion_table_nk"
        table_z_path = "/entry/dispersion_z/dispersion_table_n"
        lbda = h5file[f"{table_x_path}/wavelength"][()]

        table_x = Table(lbda=lbda, n=h5file[f"{table_x_path}/refractive_index"][()])
        table_y = Table(
            lbda=h5file[f"{table_y_path}/wavelength"][()],
            n=h5file[f"{table_y_path}/refractive_index"][()],
        )
        table_z = Table(
            lbda=h5file[f"{table_z_path}/wavelength"][()],
            n=h5file[f"{table_z_path}/refractive_index"][()],
        )

    nexus_material = read_nexus_materials(nexus)["entry"]

    assert isinstance(nexus_material, BiaxialMaterial)

    assert_array_almost_equal(
        table_x.get_dielectric(lbda), nexus_material.dispersion_x.get_dielectric(lbda)
    )

    assert_array_almost_equal(
        table_y.get_dielectric(lbda), nexus_material.dispersion_y.get_dielectric(lbda)
    )

    assert_array_almost_equal(
        table_z.get_dielectric(lbda), nexus_material.dispersion_z.get_dielectric(lbda)
    )


def test_read_nexus_dispersion_sum(datadir):
    """A dispersion sum is read correctly"""
    nexus = datadir / "YVO4-Shi-e-20C.nxs"
    lbda = np.linspace(480, 1340, 100)

    def read_sellmeier(h5file: h5py.File):
        sellmeier_path = "/entry/dispersion_x/sellmeier"
        params_A = h5file[f"{sellmeier_path}/A/values"][()]
        params_B = h5file[f"{sellmeier_path}/B/values"][()]
        params_e1 = h5file[f"{sellmeier_path}/e1/values"][()]
        params_e2 = h5file[f"{sellmeier_path}/e2/values"][()]
        eps_inf = h5file["/entry/dispersion_x/sellmeier/eps_inf/value"][()]

        sellmeier = SellmeierCustomExponent()
        for A, B, e1, e2 in zip(params_A, params_B, params_e1, params_e2):
            sellmeier.add(A, e1, B, e2)

        return sellmeier + eps_inf

    def read_poly(h5file: h5py.File):
        polynomial_path = "/entry/dispersion_x/polynomial"
        params_e = h5file[f"{polynomial_path}/e/values"][()]
        params_f = h5file[f"{polynomial_path}/f/values"][()]

        poly = Polynomial(0)
        for e, f in zip(params_e, params_f):
            poly.add(f / 1e3**e, e)

        return poly

    with h5py.File(nexus, "r") as h5file:
        disp = read_sellmeier(h5file) + read_poly(h5file)

    nexus_material = read_nexus_materials(nexus)["entry"]

    assert isinstance(nexus_material, IsotropicMaterial)

    assert_array_almost_equal(
        disp.get_dielectric(lbda), nexus_material.dispersion_x.get_dielectric(lbda)
    )



# Encoding: utf-8
"""Fractional dimensional Tanguy model."""

import numpy as np
import numpy.typing as npt
from numpy.lib.scimath import sqrt

# pylint: disable=no-name-in-module
from scipy.special import digamma, gamma

from ..utils import conversion_wavelength_energy
from .base_dispersion import Dispersion


class Tanguy(Dispersion):
    r"""Fractional dimensional Tanguy model.
    This model is an analytical expression of Wannier excitons, including
    bound and unbound states.

    Single parameters:
          :A: Amplitude (eV). Defaults to 1.
          :d: Dimensionality 1 < d <= 3. Defaults to 3.
          :gamma: Excitonic broadening (eV). Defaults to 0.1.
          :R: Excitonic binding energy (eV). Defaults to 0.1.
          :Eg: Optical band gap energy (eV). Defaults to 1.
          :a: Sellmeier coefficient for background dielectric constant (eV²).
            Defaults to 0.
          :b: Sellmeier coefficient for background dielectric constant (eV²).
            Defaults to 0.

    Repeated parameters.
        --

    Output:
        The Tanguy dispersion. Since the formula is rather long it is not written here.
        Please refer to the references for a full formula.

    References:
        * C. Tanguy, Phys. Rev. Lett. 75, 4090 (1995). Errata, Phys. Rev. Lett. 76, 716 (1996).
        * C. Tanguy, Phys. Rev. B. 60. 10660 (1990).
    """

    single_params_template = {
        "A": 1,
        "d": 3,
        "gamma": 0.1,
        "R": 0.1,
        "Eg": 1,
        "a": 0,
        "b": 0,
    }
    rep_params_template = {}

    def dielectric_function(self, lbda: npt.ArrayLike) -> npt.NDArray:
        E = conversion_wavelength_energy(lbda)
        A = self.single_params.get("A")
        d = self.single_params.get("d")
        gam = self.single_params.get("gamma")
        R = self.single_params.get("R")
        Eg = self.single_params.get("Eg")
        a = self.single_params.get("a")
        b = self.single_params.get("b")

        return (
            1
            + a / (b - E**2)
            + A
            * R ** (d / 2 - 1)
            / (E + 1j * gam) ** 2
            * (
                Tanguy.g(Tanguy.xsi(E + 1j * gam, R, Eg), d)
                + Tanguy.g(Tanguy.xsi(-E - 1j * gam, R, Eg), d)
                - 2 * Tanguy.g(Tanguy.xsi(E * 0, R, Eg), d)
            )
        )

    @staticmethod
    def xsi(z, R, Eg):
        return sqrt(R / (Eg - z))

    @staticmethod
    def g(xsi, d):
        if d == 2:
            return 2 * np.log(xsi) - 2 * digamma(0.5 - xsi)
        if d == 3:
            return 2 * np.log(xsi) - 2 * digamma(1 - xsi) - 1 / xsi

        D = d - 1
        return (
            2
            * np.pi
            * gamma(D / 2 + xsi)
            / gamma(D / 2) ** 2
            / gamma(1 - D / 2 + xsi)
            / xsi ** (d - 2)
            * (1 / np.tan(np.pi * (D / 2 - xsi)) - 1 / np.tan(np.pi * D))
        )



#!/usr/bin/python
# encoding: utf-8
import warnings

import elli
import numpy as np
from pytest import raises


def test_solver2x2_active_medium():
    s = elli.Structure(elli.AIR, [], elli.ConstantRefractiveIndex(2 - 1j).get_mat())
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        s.evaluate([200, 300, 400, 500], 70, solver=elli.Solver2x2)
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)


def test_solver2x2_inverse_active_medium():
    s = elli.Structure(elli.AIR, [], elli.ConstantRefractiveIndex(-2 + 1j).get_mat())
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        s.evaluate([200, 300, 400, 500], 70, solver=elli.Solver2x2)
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)



"""
Mueller matrix
==============

Mueller matrix fit to a SiO2 on Si measurement.
"""

# %%
import elli
from elli.fitting import ParamsHist, fit_mueller_matrix

# sphinx_gallery_thumbnail_path = '_static/mueller_matrix.png'


# %%
# Read data
# ---------
#
# We load the data from an ascii file containing each of the mueller matrix elements.
# The wavelength range is cut to be in between 210 nm and 820 nm,
# to stay in the range of the provided literature values for Si.
# The data is expected to be in a pandas dataframe containing the columns Mxy,
# where x and y refer to the matrix element inside the mueller matrix.
# The data is scaled by the M11 element, such that :math:`M_{11} = 1` for all wavelengths.
# To show the structure we print the `MM` dataframe.
# If you load your data from another source make sure it adheres to this form.
MM = elli.read_spectraray_mmatrix("Wafer_MM_70.txt").loc[210:820]
print(MM)

# %%
# Setting start parameters
# ------------------------
# Here we set the start parameters for the SiO2 cauchy dispersion
# and thickness of the layer.
params = ParamsHist()
params.add("SiO2_n0", value=1.452, min=-100, max=100, vary=True)
params.add("SiO2_n1", value=36.0, min=-40000, max=40000, vary=True)
params.add("SiO2_n2", value=0, min=-40000, max=40000, vary=True)
params.add("SiO2_k0", value=0, min=-100, max=100, vary=True)
params.add("SiO2_k1", value=0, min=-40000, max=40000, vary=True)
params.add("SiO2_k2", value=0, min=-40000, max=40000, vary=True)
params.add("SiO2_d", value=120, min=0, max=40000, vary=True)

# %%
# Load silicon dispersion from the refractiveindexinfo database
# -------------------------------------------------------------
# You can load any material from the index
# `refractiveindex.info <https://refractiveindex.info>`__, which is
# embedded into the software (so you may use it offline, too). Here, we
# are interested in the literature values for the silicon substrate.
# First we need to load the database with ``rii_db = elli.db.RII()`` and
# then we can query it with ``rii_db.get_mat("Si", "Aspnes")`` to load
# this
# `entry <https://refractiveindex.info/?shelf=main&book=Si&page=Aspnes>`__.
rii_db = elli.db.RII()
Si = rii_db.get_mat("Si", "Aspnes")


# %%
# Building the model
# ------------------
# Here the model is build and the experimental structure is returned.
# For details on this process please refer to the :ref:`Basic usage` example.
# When executed in an jupyter notebook this displays an interactive graph
# with which you can select the start parameters before fitting the data.
@fit_mueller_matrix(MM, params, display_single=False, sharex=True, full_scale=False)
def model(lbda, params):
    SiO2 = elli.Cauchy(
        params["SiO2_n0"],
        params["SiO2_n1"],
        params["SiO2_n2"],
        params["SiO2_k0"],
        params["SiO2_k1"],
        params["SiO2_k2"],
    ).get_mat()

    Layer = [elli.Layer(SiO2, params["SiO2_d"])]

    return elli.Structure(elli.AIR, Layer, Si).evaluate(
        lbda, 70, solver=elli.Solver4x4, propagator=elli.PropagatorExpm()
    )


# %%
# Plot & Fit the model
# --------------------
# Here we plot the model at the initial parameter set vs. the experimental data.
model.plot()

# %%
# We can also plot the residual between measurement and model.
model.plot_residual()

# %%
# Now we execute a fit and plot the model afterwards.
fit_stats = model.fit()
model.plot(full_scale=False)

# %%
# For comparison we plot the residual again to have a figure of merit
# for the fit quality
model.plot_residual()

# %%
# We may also print the fit statistics.
fit_stats

# %%
# References
# ----------
# `Here <https://github.com/PyEllips/pyElli/tree/master/examples/SiO2_Si%20Mueller%20Matrix>`_
# you can find the latest jupyter notebook and data files of this example.



"""Tests for a TiO2/SiO2/Si reference layer"""

from __future__ import unicode_literals

import pytest

from fixtures import datadir  # pylint: disable=unused-import
import elli


@pytest.mark.parametrize(
    "filename",
    ["ellips.test.nxs", "ellips_nx_opt.test.nxs"],
)
# pylint: disable=redefined-outer-name
def test_reading_of_psi_delta_nxs(datadir, filename):
    """Psi/delta NeXus file is read w/o errors"""
    elli.read_nexus_psi_delta(datadir / filename)


@pytest.mark.parametrize(
    "filename",
    ["ellips.test.nxs", "ellips_nx_opt.test.nxs"],
)
# pylint: disable=redefined-outer-name
def test_reading_and_conv_to_rho(datadir, filename):
    """Rho values are read from Psi / Delta file"""
    elli.read_nexus_rho(datadir / filename)



#!/usr/bin/python
# encoding: utf-8

import elli
import numpy as np


class TestUniaxial:
    n_i = 1.0  # incident medium is air
    n_o = 2.0  # ordinary index of thin layer
    n_e = 2.5  # extraordinary index of thin layer
    lbda = 620  # Wavelength for evaluation (nm)
    Phi_i = 70  # 70° incidence angle (degree)self.
    d = 100  # thin layer thickness (nm)
    Phi_E = 45  # 1st Euler angle
    Theta_E = 45  # 2nd Euler angle

    filmMaterial = elli.UniaxialMaterial(
        elli.ConstantRefractiveIndex(n=n_o), elli.ConstantRefractiveIndex(n=n_e)
    )
    R = elli.rotation_euler(Phi_E, Theta_E, 0)
    filmMaterial.set_rotation(R)

    air = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_i))
    Kx = n_i * np.sin(np.deg2rad(Phi_i))
    film = elli.Layer(filmMaterial, d)
    epsilon = filmMaterial.get_tensor(lbda)
    Delta = elli.Solver4x4.build_delta_matrix(Kx, epsilon)

    air = elli.AIR
    epsilon_air = air.get_tensor(lbda)

    Tp = elli.PropagatorExpm().calculate_propagation(Delta, -d, np.array([lbda]))
    Li = elli.Solver4x4.transition_matrix_iso_halfspace(
        np.array([Kx]), epsilon_air, inv=True
    )

    n_t = 3.898 + 0.016j  # refractive index of substrate
    silicon = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n=n_t))
    epsilon_silicon = silicon.get_tensor(lbda)
    Lt = elli.Solver4x4.transition_matrix_iso_halfspace(
        np.array([Kx]), epsilon_silicon, inv=False
    )
    s = elli.Structure(air, [film], silicon)
    data = s.evaluate(np.array([lbda]), Phi_i)

    def test_rotated_tensor(self):
        tensor = np.array(
            [
                [
                    [4.5625 + 0j, -0.5625 + 0j, 0.7955 + 0j],
                    [-0.5625 + 0j, 4.5625 + 0j, -0.7955 + 0j],
                    [0.7955 + 0j, -0.7955 + 0j, 5.125 + 0j],
                ]
            ]
        )

        np.testing.assert_array_almost_equal(
            self.filmMaterial.get_tensor(self.lbda), tensor, decimal=4
        )

    def test_build_delta_matrix(self):
        delta_matrix = np.array(
            [
                [-0.1459, 0.1459, 0.0, 0.8277],
                [0.0, 0.0, -1.0, 0.0],
                [0.439, -3.556, 0.0, -0.1459],
                [4.439, -0.439, 0.0, -0.1459],
            ],
            dtype=np.complex128,
        )

        np.testing.assert_array_almost_equal(self.Delta[0], delta_matrix, decimal=3)

    def test_transition_matrix(self):
        transition_matrix = np.array(
            [
                [
                    [-0.352 - 0.057j, 0.091 - 0.001j, 0.033 + 0.044j, 0.056 - 0.397j],
                    [0.104 + 0.08j, -0.326 - 0.011j, 0.004 + 0.502j, -0.033 - 0.044j],
                    [0.162 - 0.014j, -0.017 + 1.753j, -0.326 - 0.011j, -0.091 + 0.001j],
                    [0.300 - 2.12j, -0.162 + 0.014j, -0.104 - 0.08j, -0.352 - 0.057j],
                ]
            ]
        )

        np.testing.assert_array_almost_equal(self.Tp, transition_matrix, decimal=3)

    def test_li_matrix(self):
        n_i = 1
        cos_phi = 0.342

        li = (
            np.array(
                [
                    [
                        [0, 1, -1 / n_i / cos_phi, 0],
                        [0, 1, 1 / n_i / cos_phi, 0],
                        [1 / cos_phi, 0, 0, 1 / n_i],
                        [-1 / cos_phi, 0, 0, 1 / n_i],
                    ]
                ],
                dtype=np.complex128,
            )
            * 0.5
        )

        np.testing.assert_array_almost_equal(self.Li, li, decimal=3)

    def test_lt_matrix(self):
        n_x = 3.898 + 1j * 0.016
        cos_phi = 0.9705 + 1j * 2.4e-4
        lt = np.array(
            [
                [
                    [0, 0, cos_phi, -cos_phi],
                    [1, 1, 0, 0],
                    [-n_x * cos_phi, n_x * cos_phi, 0, 0],
                    [0, 0, n_x, n_x],
                ]
            ],
            dtype=np.complex128,
        )

        np.testing.assert_array_almost_equal(self.Lt, lt, decimal=3)

    def test_transition_matrix_with_halfspace(self):
        # Fujiwara Original
        tmatrix = np.array(
            [
                [
                    [
                        -1.946 - 1j * 3.588,
                        1.668 - 1j * 1.548,
                        0.273 + 1j * 0.03,
                        0.63 - 1j * 0.148,
                    ],
                    [
                        1.614 + 1j * 1.679,
                        -1.989 + 1j * 3.435,
                        -0.301 - 1j * 0.064,
                        -0.861 - 1j * 0.101,
                    ],
                    [
                        0.065 - 1j * 0.086,
                        0.039 + 1j * 0.097,
                        -0.71 - 1j * 3.486,
                        -0.003 - 1j * 1.266,
                    ],
                    [
                        0.167 + 1j * 0.403,
                        -0.593 - 1j * 0.386,
                        -0.369 + 1j * 1.199,
                        -1.66 + 1j * 3.094,
                    ],
                ]
            ]
        )

        np.testing.assert_array_almost_equal(
            self.Li @ self.Tp @ self.Lt, tmatrix, decimal=1
        )

    def test_jones_calculation(self):
        jones_matrix = np.array(
            [[[-0.311 - 0.161j, -0.107 - 0.002j], [0.042 - 0.036j, -0.551 + 0.151j]]]
        )

        np.testing.assert_array_almost_equal(
            self.data.jones_matrix_r, jones_matrix, decimal=3
        )

    def test_psi_calculation(self):
        psi = np.array([[[31.46, 10.57], [5.54, 45.0]]])

        np.testing.assert_array_almost_equal(self.data.psi_matrix, psi, decimal=2)

    def test_delta_calculation(self):
        delta = np.array([[[-42.67, -16.35], [-154.93, 0.0]]])

        np.testing.assert_array_almost_equal(self.data.delta_matrix, delta, decimal=1)



# Encoding: utf-8
"""Constant epsilon infinity."""

from typing import Any, Dict
import numpy.typing as npt

from .base_dispersion import Dispersion


class EpsilonInf(Dispersion):
    r"""Constant epsilon infinity.

    Single parameters:
        :eps: Constant value for the constant epsilon. Defaults to 1.

    Repeated parameters:
        --

    Output:
        .. math::
            \varepsilon(\lambda) = \textbf{eps}
    """

    single_params_template = {"eps": 1}
    rep_params_template: Dict[str, Any] = {}

    def dielectric_function(self, _: npt.ArrayLike) -> npt.NDArray:
        return self.single_params.get("eps")



"""Tests for a TiO2/SiO2/Si reference layer"""

from __future__ import unicode_literals

import pytest

from fixtures import datadir  # pylint: disable=unused-import
import elli


# pylint: disable=redefined-outer-name
def test_reading_of_psi_delta_spetraray(datadir):
    """Psi/delta Spectraray file is read w/o errors"""
    elli.read_spectraray_psi_delta(datadir / "Si_SiO2_theta_50_60_70.txt")


# pylint: disable=redefined-outer-name
def test_reading_and_conv_to_spectraray(datadir):
    """Rho values are read from Psi / Delta file"""
    elli.read_spectraray_rho(datadir / "Si_SiO2_theta_50_60_70.txt")



"""
Interface reflection
=====================
"""

# %%
# Interface between two materials
# -------------------------------
# Authors: O. Castany, C. Molinaro, M. Müller
#
# Interface between two materials n1/n2.
# Calculations of the transmission and reflexion coefficients with varying incidence angle.
import elli
import matplotlib.pyplot as plt
import numpy as np
from scipy.constants import c, pi

# %%
# Structure definition
# ------------------------
n1 = 1
n2 = 1.5
front = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n1))
back = elli.IsotropicMaterial(elli.ConstantRefractiveIndex(n2))

# Structure
s = elli.Structure(front, [], back)

# Parameters for the calculation
lbda = 1000
k0 = 2 * pi / lbda
Phi_list = np.linspace(0, 89, 90)  #  range for the incidence angles

# %%
# Analytical calculation
# ------------------------
Phi_i = np.deg2rad(Phi_list)

Phi_t = np.arcsin((n1 * np.sin(Phi_i) / n2).astype(complex))
kz1 = n1 * k0 * np.cos(Phi_i)
kz2 = n2 * k0 * np.cos(Phi_t)
r_s = (kz1 - kz2) / (kz1 + kz2)
t_s = 1 + r_s
r_p = (kz1 * n2**2 - kz2 * n1**2) / (kz1 * n2**2 + kz2 * n1**2)
t_p = np.cos(Phi_i) * (1 - r_p) / np.cos(Phi_t)

# Reflection and transmission coefficients, polarisation s and p
R_th_ss = abs(r_s) ** 2
R_th_pp = abs(r_p) ** 2
t2_th_ss = abs(t_s) ** 2
t2_th_pp = abs(t_p) ** 2
# The power transmission coefficient is T = Re(kz2/kz1) × |t|^2
correction = np.real(kz2 / kz1)
T_th_ss = correction * t2_th_ss
T_th_pp = correction * t2_th_pp


# %%
# Calculation with pyElli
# -----------------------
data = elli.ResultList([s.evaluate(lbda, Phi_i) for Phi_i in Phi_list])

R_pp = data.R_pp
R_ss = data.R_ss

T_pp = data.T_pp
T_ss = data.T_ss

t2_pp = np.abs(data.t_pp) ** 2
t2_ss = np.abs(data.t_ss) ** 2

# %%
# Plotting
# ----------
fig = plt.figure(figsize=(12.0, 6.0))
plt.rcParams["axes.prop_cycle"] = plt.cycler("color", "bgrcmk")
ax = fig.add_axes([0.1, 0.1, 0.7, 0.8])

d = np.vstack((R_ss, R_pp, t2_ss, t2_pp, T_ss, T_pp)).T
lines1 = ax.plot(Phi_list, d)
legend1 = ("R_ss", "R_pp", "t2_ss", "t2_pp", "T_ss", "T_pp")

d = np.vstack((R_th_ss, R_th_pp, t2_th_ss, t2_th_pp, T_th_ss, T_th_pp)).T
lines2 = ax.plot(Phi_list, d, ".")
legend2 = ("R_th_ss", "R_th_pp", "t2_th_ss", "t2_th_pp", "T_th_ss", "T_th_pp")

ax.legend(
    lines1 + lines2,
    legend1 + legend2,
    loc="upper left",
    bbox_to_anchor=(1.05, 1),
    borderaxespad=0.0,
)

ax.set_title("Interface n$_1$={:} / n$_2$={:}".format(n1, n2))
ax.set_xlabel(r"Incidence angle $\Phi_i$ ")
ax.set_ylabel(r"Reflexion and transmission coefficients $R$, $T$, $|t|^2$")

plt.show()

# %%



"""Tests for the result class"""

import numpy as np
import elli
from pytest import raises


class TestStructures:
    disp = elli.ConstantRefractiveIndex(2)
    mat = elli.IsotropicMaterial(disp)
    layer = elli.Layer(mat, 20)

    def test_layer_typeguard(self):
        """Test layer input checks."""
        with raises(TypeError):
            elli.Layer(self.disp, 20)

        with raises(ValueError):
            elli.Layer(self.mat, -1)

        with raises(TypeError):
            elli.RepeatedLayers(self.layer, 2)

        with raises(TypeError):
            elli.RepeatedLayers([self.mat], 2)

        with raises(ValueError):
            elli.RepeatedLayers([self.layer], -2)

        with raises(ValueError):
            elli.RepeatedLayers([self.layer], 2, -2)

        with raises(ValueError):
            elli.RepeatedLayers([self.layer], 2, 0, -1)

        with raises(TypeError):
            elli.TwistedLayer(self.disp, 20, 5, 20)

        with raises(ValueError):
            elli.TwistedLayer(self.mat, -1, 5, 20)

        with raises(ValueError):
            elli.TwistedLayer(self.mat, 20, 0, 20)

        with raises(TypeError):
            elli.VaryingMixtureLayer(self.disp, 20, 5)

        with raises(TypeError):
            elli.VaryingMixtureLayer(self.mat, 20, 5)

    def test_structure_typeguard(self):
        """Test layer input checks."""
        with raises(TypeError):
            elli.Structure(self.disp, [], self.mat)

        with raises(TypeError):
            elli.Structure(self.mat, [], self.disp)

        with raises(TypeError):
            elli.Structure(elli.AIR, self.layer, self.mat)

        with raises(TypeError):
            elli.Structure(elli.AIR, [self.mat], self.mat)

    def test_varying_mixture_layer(self):
        """Tests the basic functionality of the VML."""
        vca_mat = elli.VCAMaterial(elli.AIR, self.mat, 0.5)
        vml = elli.VaryingMixtureLayer(vca_mat, 10, 3)

        np.testing.assert_array_equal(
            vml.get_permittivity_profile(500)[1][1],
            (elli.AIR.get_tensor(500) + self.mat.get_tensor(500)) / 2,
        )
